from __future__ import unicode_literals
import unittest
from defcon import Contour, Font, Glyph, Point
from defcon.test.testTools import getTestFontPath


def simpleSegment(segment):
    return [(i.x, i.y, i.segmentType) for i in segment]


class ContourTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def setUp(self):
        self.font = Font()
        self.glyph = self.font.newGlyph("A")
        self.contour = Contour()

    def tearDown(self):
        del self.contour
        del self.glyph
        del self.font

    def test_getParent(self):
        self.assertIsNone(self.contour.getParent())
        self.contour = Contour(self.glyph)
        self.assertEqual(self.contour.getParent(), self.glyph)

    def test_font(self):
        self.assertIsNone(self.contour.font)
        self.contour = Contour(self.glyph)
        self.assertEqual(self.contour.font, self.font)
        with self.assertRaises(AttributeError):
            self.contour.font = "foo"

    def test_layerSet(self):
        self.assertIsNone(self.contour.layerSet)
        self.contour = Contour(self.glyph)
        self.assertIsNotNone(self.contour.layerSet)
        self.assertEqual(self.contour.layerSet, self.glyph.layerSet)
        with self.assertRaises(AttributeError):
            self.contour.layerSet = "foo"

    def test_layer(self):
        self.assertIsNone(self.contour.layer)
        self.contour = Contour(self.glyph)
        self.assertIsNotNone(self.contour.layer)
        self.assertEqual(self.contour.layer, self.glyph.layer)
        with self.assertRaises(AttributeError):
            self.contour.layer = "foo"

    def test_glyph(self):
        self.assertIsNone(self.contour.glyph)
        self.contour = Contour(self.glyph)
        self.assertEqual(self.contour.glyph, self.glyph)
        glyph = Glyph()
        self.contour = Contour()
        self.contour.glyph = glyph
        self.assertEqual(self.contour.glyph, glyph)
        with self.assertRaises(AssertionError):
            self.contour.glyph = self.glyph

    def test_list_behavior(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        contour = glyph[0]
        with self.assertRaises(IndexError):
            contour[len(contour) + 1]
        self.assertEqual(contour[len(contour):len(contour)], [])
        self.assertEqual(contour[len(contour) + 1:len(contour) + 2], [])
        self.assertNotEqual(contour[0:len(contour) + 1], [])
        self.assertEqual([(point.x, point.y) for point in contour[0:]],
                         [(point.x, point.y)
                          for point in contour[0:len(contour) + 1]])
        self.assertEqual([(point.x, point.y) for point in contour[0:]],
                         [(0, 0), (700, 0), (700, 700), (0, 700)])
        self.assertEqual([(point.x, point.y) for point in contour],
                         [(0, 0), (700, 0), (700, 700), (0, 700)])

    def test_onCurvePoints(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        contour = glyph[0]
        self.assertEqual(len(contour.onCurvePoints), 4)
        self.assertEqual([(point.x, point.y)
                          for point in contour.onCurvePoints],
                         [(0, 0), (700, 0), (700, 700), (0, 700)])
        glyph = font["B"]
        contour = glyph[0]
        self.assertEqual(len(contour.onCurvePoints), 4)
        self.assertEqual([(point.x, point.y)
                          for point in contour.onCurvePoints],
                         [(0, 350), (350, 0), (700, 350), (350, 700)])

    def test_appendPoint(self):
        pointA = Point((0, 5))
        self.assertFalse(self.contour.dirty)
        self.contour.appendPoint(pointA)
        self.assertTrue(self.contour.dirty)
        self.assertEqual([(point.x, point.y)
                          for point in self.contour],
                         [(0, 5)])
        pointB = Point((6, 7))
        self.contour.appendPoint(pointB)
        self.assertEqual([(point.x, point.y)
                          for point in self.contour],
                         [(0, 5), (6, 7)])

    def test_insertPoint(self):
        pointA = Point((0, 5))
        pointB = Point((6, 7))
        pointC = Point((8, 9))
        self.assertFalse(self.contour.dirty)
        self.contour.insertPoint(0, pointA)
        self.assertTrue(self.contour.dirty)
        self.assertEqual([(point.x, point.y)
                          for point in self.contour],
                         [(0, 5)])
        self.contour.insertPoint(0, pointB)
        self.assertEqual([(point.x, point.y)
                          for point in self.contour],
                         [(6, 7), (0, 5)])
        self.contour.insertPoint(1, pointC)
        self.assertEqual([(point.x, point.y)
                          for point in self.contour],
                         [(6, 7), (8, 9), (0, 5)])

    def test_removePoint(self):
        font = Font(getTestFontPath())
        contour = font["A"][0]
        self.assertFalse(contour.dirty)
        contour.removePoint(contour[1])
        self.assertTrue(contour.dirty)
        self.assertEqual([(point.x, point.y)
                          for point in contour],
                         [(0, 0), (700, 700), (0, 700)])
        contour.removePoint(contour[0])
        self.assertEqual([(point.x, point.y)
                          for point in contour],
                         [(700, 700), (0, 700)])
        contour.removePoint(contour[1])
        self.assertEqual([(point.x, point.y)
                          for point in contour],
                         [(700, 700)])

    def test_setStartPoint(self):
        font = Font(getTestFontPath())
        contour = font["B"][0]
        start = [(point.segmentType, point.x, point.y) for point in contour]
        contour.setStartPoint(6)
        self.assertTrue(contour.dirty)
        contour.setStartPoint(6)
        end = [(point.segmentType, point.x, point.y) for point in contour]
        self.assertEqual(start, end)
        contour = font["A"][0]
        start = [(point.segmentType, point.x, point.y) for point in contour]
        contour.setStartPoint(2)
        contour.setStartPoint(2)
        end = [(point.segmentType, point.x, point.y) for point in contour]
        self.assertEqual(start, end)
        contour = font["B"][0]
        start = [(point.segmentType, point.x, point.y) for point in contour]
        contour.setStartPoint(3)
        contour.setStartPoint(9)
        end = [(point.segmentType, point.x, point.y) for point in contour]
        self.assertEqual(start, end)

    def test_len(self):
        font = Font(getTestFontPath())
        contour = font["A"][0]
        self.assertEqual(len(contour), 4)
        contour = font["B"][0]
        self.assertEqual(len(contour), 12)

    def test_iter(self):
        font = Font(getTestFontPath())
        contour = font["A"][0]
        self.assertEqual([(point.x, point.y) for point in contour],
                         [(0, 0), (700, 0), (700, 700), (0, 700)])

    def test_index(self):
        font = Font(getTestFontPath())
        contour = font["B"][0]
        self.assertEqual(contour.index(contour[2]), 2)

    def test_reverse(self):
        font = Font(getTestFontPath())
        contour = font["A"][0]
        contour.reverse()
        self.assertEqual([(point.x, point.y) for point in contour._points],
                         [(0, 0), (0, 700), (700, 700), (700, 0)])
        contour.reverse()
        self.assertEqual([(point.x, point.y) for point in contour._points],
                         [(0, 0), (700, 0), (700, 700), (0, 700)])

    def test_segments(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        contour = glyph[0]
        self.assertEqual(
            [simpleSegment(segment) for segment in contour.segments],
            [[(700, 0, "line")], [(700, 700, "line")],
             [(0, 700, "line")], [(0, 0, "line")]])
        glyph = font["B"]
        contour = glyph[0]
        self.assertEqual(
            [simpleSegment(segment) for segment in contour.segments],
            [[(0, 157, None), (157, 0, None), (350, 0, "curve")],
             [(543, 0, None), (700, 157, None), (700, 350, "curve")],
             [(700, 543, None), (543, 700, None), (350, 700, "curve")],
             [(157, 700, None), (0, 543, None), (0, 350, "curve")]])

    def test_removeSegment_lines(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        contour = glyph[0]
        contour.removeSegment(len(contour.segments) - 2)
        self.assertEqual(
            [simpleSegment(segment) for segment in contour.segments],
            [[(700, 0, "line")], [(700, 700, "line")],
             [(0, 0, "line")]])
        contour.removeSegment(len(contour.segments) - 1)
        self.assertEqual(
            [simpleSegment(segment) for segment in contour.segments],
            [[(700, 700, "line")], [(700, 0, "line")]])

    def test_removeSegment_first_segment(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        contour = glyph[0]
        contour.removeSegment(0)
        self.assertEqual(
            [simpleSegment(segment) for segment in contour.segments],
            [[(700, 700, "line")], [(0, 700, "line")],
             [(0, 0, "line")]])

    def test_removeSegment_last_segment(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        contour = glyph[0]
        contour.removeSegment(len(contour.segments) - 1)
        self.assertEqual(
            [simpleSegment(segment) for segment in contour.segments],
            [[(700, 700, "line")], [(0, 700, "line")],
             [(700, 0, "line")]])

    def test_removeSegment_curves(self):
        font = Font(getTestFontPath())
        glyph = font["B"]
        contour = glyph[0]
        contour.removeSegment(len(contour.segments) - 2)
        self.assertEqual(
            [simpleSegment(segment) for segment in contour.segments],
            [[(0, 157, None), (157, 0, None), (350, 0, "curve")],
             [(543, 0, None), (700, 157, None), (700, 350, "curve")],
             [(157, 700, None), (0, 543, None), (0, 350, "curve")]])

    def test_removeSegment_curves_preserveCurve(self):
        font = Font(getTestFontPath())
        glyph = font["B"]
        contour = glyph[0]
        contour.removeSegment(len(contour.segments) - 2, preserveCurve=True)
        self.assertEqual(
            [simpleSegment(segment) for segment in contour.segments],
            [[(0, 157, None), (157, 0, None), (350, 0, "curve")],
             [(543, 0, None), (700, 157, None), (700, 350, "curve")],
             [(700.0, 736.0, None), (0.0, 736.0, None), (0, 350, "curve")]])

    def test_offCurveOnlySegement(self):
        font = Font()
        glyph = font.newGlyph("offCurveOnly")
        pen = glyph.getPen()
        pen.qCurveTo((100, 100), (100, 200), (200, 200), (200, 100), None)
        pen.endPath()
        contour = glyph[0]
        self.assertEqual(
            [simpleSegment(segment) for segment in contour.segments],
            [[(100, 100, None), (100, 200, None), (200, 200, None), (200, 100, None)]]
        )

    def test_clockwise_get(self):
        font = Font(getTestFontPath())
        contour = font["A"][0]
        self.assertFalse(contour.clockwise)
        contour = font["A"][1]
        self.assertTrue(contour.clockwise)
        contour._clockwiseCache = None
        contour.clockwise = False
        self.assertFalse(contour.clockwise)
        contour._clockwiseCache = None
        contour.clockwise = True
        self.assertTrue(contour.clockwise)

    def test_clockwise_set(self):
        font = Font(getTestFontPath())
        contour = font["A"][0]
        contour.clockwise = False
        self.assertFalse(contour.clockwise)
        contour._clockwiseCache = None
        contour.clockwise = True
        self.assertTrue(contour.clockwise)

    def test_open(self):
        font = Font(getTestFontPath("TestOpenContour.ufo"))
        glyph = font["A"]
        self.assertTrue(glyph[0].open)
        self.assertFalse(glyph[1].open)
        self.assertTrue(glyph[2].open)
        self.assertFalse(glyph[3].open)
        contour = Contour()
        self.assertTrue(contour.open)

    def test_bounds(self):
        font = Font(getTestFontPath())
        contour = font["A"][0]
        self.assertEqual(contour.bounds, (0, 0, 700, 700))

    def test_controlPointBounds(self):
        font = Font(getTestFontPath())
        contour = font["A"][0]
        self.assertEqual(contour.controlPointBounds, (0, 0, 700, 700))

    def test_move(self):
        font = Font(getTestFontPath())
        contour = font["A"][0]
        contour.move((100, 100))
        self.assertEqual(contour.bounds, (100, 100, 800, 800))
        self.assertTrue(contour.dirty)

    def test_pointInside(self):
        font = Font(getTestFontPath())
        contour = font["A"][0]
        self.assertTrue(contour.pointInside((100, 100)))
        self.assertFalse(contour.pointInside((0, 0)))
        self.assertFalse(contour.pointInside((-100, -100)))

    def test_positionForProspectivePointInsertionAtSegmentAndT(self):
        font = Font(getTestFontPath())
        contour = font["A"][0]
        self.assertEqual(
            contour.positionForProspectivePointInsertionAtSegmentAndT(0, .5),
            ((350.0, 0.0), False))
        contour = font["B"][0]
        self.assertEqual(
            contour.positionForProspectivePointInsertionAtSegmentAndT(0, .5),
            ((102.625, 102.625), True))
        contour = font["B"][1]
        self.assertEqual(
            contour.positionForProspectivePointInsertionAtSegmentAndT(0, .5),
            ((226.125, 473.5), True))

    def test_splitAndInsertPointAtSegmentAndT(self):
        font = Font(getTestFontPath())
        contour = font["A"][0]
        contour.splitAndInsertPointAtSegmentAndT(0, .5)
        self.assertEqual(
            [(point.x, point.y, point.segmentType) for point in contour],
            [(0, 0, "line"), (350.0, 0.0, "line"), (700, 0, "line"),
             (700, 700, "line"), (0, 700, "line")])
        contour = font["B"][0]
        contour.splitAndInsertPointAtSegmentAndT(0, .5)
        self.assertEqual(
            [(point.x, point.y, point.segmentType) for point in contour],
            [(0, 350, "curve"), (0.0, 253.5, None), (39.25, 166.0, None),
             (102.625, 102.625, "curve"), (166.0, 39.25, None),
             (253.5, 0.0, None), (350, 0, "curve"), (543, 0, None),
             (700, 157, None), (700, 350, "curve"), (700, 543, None),
             (543, 700, None), (350, 700, "curve"), (157, 700, None),
             (0, 543, None)])

    def test_identifier(self):
        glyph = Glyph()
        contour = Contour()
        glyph.appendContour(contour)
        contour.identifier = "contour 1"
        self.assertEqual(contour.identifier, "contour 1")
        self.assertEqual(sorted(glyph.identifiers), ["contour 1"])
        contour = Contour()
        glyph.appendContour(contour)
        with self.assertRaises(AssertionError):
            contour.identifier = "contour 1"
        contour.identifier = "contour 2"
        self.assertEqual(sorted(glyph.identifiers), ["contour 1", "contour 2"])
        contour.identifier = "not contour 2 anymore"
        self.assertEqual(contour.identifier, "contour 2")
        self.assertEqual(sorted(glyph.identifiers), ["contour 1", "contour 2"])
        contour.identifier = None
        self.assertEqual(contour.identifier, "contour 2")
        self.assertEqual(sorted(glyph.identifiers), ["contour 1", "contour 2"])

    def test_correct_direction_same_area(self):
        glyph = Glyph()
        pen = glyph.getPointPen()
        pen.beginPath()
        pen.addPoint((0, 0), segmentType="line")
        pen.addPoint((0, 50), segmentType="line")
        pen.addPoint((50, 50), segmentType="line")
        pen.endPath()
        pen.beginPath()
        pen.addPoint((50, 50), segmentType="line")
        pen.addPoint((50, 100), segmentType="line")
        pen.addPoint((100, 100), segmentType="line")
        pen.endPath()
        try:
            glyph.correctContourDirection()
        except Exception as e:
            self.fail("glyph.correctContourDirection() raised unexpected exception: "
                      + str(e))

if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import
from defcon.pens.glyphObjectPointPen import GlyphObjectPointPen
from defcon.pens.transformPointPen import TransformPointPen
from defcon.objects.component import _defaultTransformation


class DecomposeComponentPointPen(GlyphObjectPointPen):

    def __init__(self, glyph, layer):
        self._layer = layer
        super(DecomposeComponentPointPen, self).__init__(glyph)

    def addComponent(self, baseGlyphName, transformation, identifier=None, **kwargs):
        if baseGlyphName in self._layer:
            baseGlyph = self._layer[baseGlyphName]
            if transformation == _defaultTransformation:
                baseGlyph.drawPoints(self)
            else:
                transformPointPen = TransformPointPen(self, transformation)
                baseGlyph.drawPoints(transformPointPen)



import unittest
from defcon import Font, Glyph, Contour, Component, Anchor, Guideline, Layer
from defcon.test.testTools import getTestFontPath


class GlyphTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_identifiers(self):
        glyph = Glyph()
        pointPen = glyph.getPointPen()
        pointPen.beginPath(identifier="contour 1")
        pointPen.addPoint((0, 0), identifier="point 1")
        pointPen.addPoint((0, 0), identifier="point 2")
        pointPen.endPath()
        pointPen.beginPath(identifier="contour 2")
        pointPen.endPath()
        pointPen.addComponent("A", (1, 1, 1, 1, 1, 1),
                              identifier="component 1")
        pointPen.addComponent("A", (1, 1, 1, 1, 1, 1),
                              identifier="component 2")
        guideline = Guideline()
        guideline.identifier = "guideline 1"
        glyph.appendGuideline(guideline)
        guideline = Guideline()
        guideline.identifier = "guideline 2"
        glyph.appendGuideline(guideline)

        self.assertEqual([contour.identifier for contour in glyph],
                         ["contour 1", "contour 2"])
        self.assertEqual([point.identifier for point in glyph[0]],
                         ["point 1", "point 2"])
        self.assertEqual(
            [component.identifier for component in glyph.components],
            ["component 1", "component 2"])
        with self.assertRaises(AssertionError):
            pointPen.beginPath(identifier="contour 1")
        pointPen.endPath()

        pointPen.beginPath()
        pointPen.addPoint((0, 0))
        with self.assertRaises(AssertionError):
            pointPen.addPoint((0, 0), identifier="point 1")
        pointPen.endPath()

        with self.assertRaises(AssertionError):
            pointPen.addComponent("A", (1, 1, 1, 1, 1, 1),
                                  identifier="component 1")

        g = Guideline()
        g.identifier = "guideline 1"
        with self.assertRaises(AssertionError):
            glyph.appendGuideline(g)

        self.assertEqual(
            sorted(glyph.identifiers),
            ["component 1", "component 2", "contour 1", "contour 2",
             "guideline 1", "guideline 2", "point 1", "point 2"])
        glyph.removeContour(glyph[0])
        self.assertEqual(
            sorted(glyph.identifiers),
            ["component 1", "component 2", "contour 2",
             "guideline 1", "guideline 2"])
        glyph.removeComponent(glyph.components[0])
        self.assertEqual(
            sorted(glyph.identifiers),
            ["component 2", "contour 2", "guideline 1", "guideline 2"])
        glyph.removeGuideline(glyph.guidelines[0])
        self.assertEqual(
            sorted(glyph.identifiers),
            ["component 2", "contour 2", "guideline 2"])

    def test_name_set(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.name = "RenamedGlyph"
        self.assertEqual(glyph.name, "RenamedGlyph")
        self.assertEqual(sorted(font.keys()), ["B", "C", "RenamedGlyph"])

        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.name = "A"
        self.assertFalse(glyph.dirty)

    def test_name_get(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(glyph.name, "A")

    def test_unicodes_get(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(glyph.unicodes, [65])

    def test_unicodes_set(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.unicodes = [123, 456]
        self.assertEqual(glyph.unicodes, [123, 456])
        self.assertTrue(glyph.dirty)

    def test_unicode_get(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(glyph.unicode, 65)

    def test_unicode_set(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.unicode = 123
        self.assertEqual(glyph.unicodes, [123])
        glyph.unicode = 456
        self.assertEqual(glyph.unicodes, [456])
        self.assertTrue(glyph.dirty)

    def test_bounds(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(glyph.bounds, (0, 0, 700, 700))
        glyph = font["B"]
        self.assertEqual(glyph.bounds, (0, 0, 700, 700))
        glyph = font["C"]
        self.assertEqual(glyph.bounds, (0.0, 0.0, 700.0, 700.0))

    def test_controlPointBounds(self):
        from defcon.test.testTools import getTestFontPath
        from defcon.objects.font import Font
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(glyph.controlPointBounds, (0, 0, 700, 700))
        glyph = font["B"]
        self.assertEqual(glyph.controlPointBounds, (0, 0, 700, 700))
        glyph = font["C"]
        self.assertEqual(glyph.controlPointBounds, (0.0, 0.0, 700.0, 700.0))

    def test_leftMargin_get(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(glyph.leftMargin, 0)
        glyph = font["B"]
        self.assertEqual(glyph.leftMargin, 0)

    def test_leftMargin_set(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.leftMargin = 100
        self.assertEqual(glyph.leftMargin, 100)
        self.assertEqual(glyph.width, 800)
        self.assertTrue(glyph.dirty)

    def test_rightMargin_get(self):
        from defcon.test.testTools import getTestFontPath
        from defcon.objects.font import Font
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(glyph.rightMargin, 0)

    def test_rightMargin_set(self):
        from defcon.test.testTools import getTestFontPath
        from defcon.objects.font import Font
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.rightMargin = 100
        self.assertEqual(glyph.rightMargin, 100)
        self.assertEqual(glyph.width, 800)
        self.assertTrue(glyph.dirty)

    def test_bottomMargin_get(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(glyph.bottomMargin, 0)
        glyph = font["B"]
        self.assertEqual(glyph.bottomMargin, 0)
        # empty glyph
        glyph = font.newGlyph("D")
        self.assertIsNone(glyph.bottomMargin)

    def test_bottomMargin_set(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.bottomMargin = 100
        self.assertEqual(glyph.bottomMargin, 100)
        self.assertEqual(glyph.height, 600)
        self.assertEqual(glyph.verticalOrigin, 500)
        self.assertTrue(glyph.dirty)
        # now glyph.verticalOrigin is defined
        glyph.bottomMargin = 50
        self.assertEqual(glyph.bottomMargin, 50)
        self.assertEqual(glyph.height, 550)
        self.assertEqual(glyph.verticalOrigin, 500)
        self.assertTrue(glyph.dirty)
        # empty glyph
        glyph = font.newGlyph("D")
        glyph.dirty = False
        glyph.bottomMargin = 10
        self.assertIsNone(glyph.bottomMargin)
        self.assertEqual(glyph.height, 0)
        self.assertIsNone(glyph.verticalOrigin)
        self.assertFalse(glyph.dirty)


    def test_topMargin_get(self):
        from defcon.test.testTools import getTestFontPath
        from defcon.objects.font import Font
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(glyph.topMargin, -200)
        # empty glyph
        glyph = font.newGlyph("D")
        self.assertIsNone(glyph.topMargin)

    def test_topMargin_set(self):
        from defcon.test.testTools import getTestFontPath
        from defcon.objects.font import Font
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.topMargin = 100
        self.assertEqual(glyph.topMargin, 100)
        self.assertEqual(glyph.height, 800)
        self.assertEqual(glyph.verticalOrigin, 800)
        self.assertTrue(glyph.dirty)
        # now glyph.verticalOrigin is defined
        glyph.topMargin = 50
        self.assertEqual(glyph.topMargin, 50)
        self.assertEqual(glyph.height, 750)
        self.assertEqual(glyph.verticalOrigin, 750)
        self.assertTrue(glyph.dirty)
        # empty glyph
        glyph = font.newGlyph("D")
        glyph.dirty = False
        glyph.topMargin = 10
        self.assertIsNone(glyph.topMargin)
        self.assertEqual(glyph.height, 0)
        self.assertIsNone(glyph.verticalOrigin)
        self.assertFalse(glyph.dirty)


    def test_width_get(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(glyph.width, 700)

    def test_width_set(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.width = 100
        self.assertEqual(glyph.width, 100)
        self.assertTrue(glyph.dirty)

    def test_height_get(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(glyph.height, 500)

    def test_height_set(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.height = 100
        self.assertEqual(glyph.height, 100)
        self.assertEqual(glyph.verticalOrigin, None)
        self.assertTrue(glyph.dirty)

    def test_markColor(self):
        from defcon.objects.font import Font
        font = Font()
        font.newGlyph("A")
        glyph = font["A"]
        self.assertIsNone(glyph.markColor)
        glyph.markColor = "1,0,1,0"
        self.assertEqual(glyph.markColor, "1,0,1,0")
        glyph.markColor = "1,0,1,0"
        self.assertEqual(glyph.markColor, "1,0,1,0")
        glyph.markColor = None
        self.assertIsNone(glyph.markColor)

    def test_verticalOrigin(self):
        from defcon.test.testTools import getTestFontPath
        from defcon.objects.font import Font
        font = Font()
        font.newGlyph("A")
        glyph = font["A"]
        self.assertIsNone(glyph.verticalOrigin)
        self.assertEqual(glyph.height, 0)
        glyph.verticalOrigin = 1000
        self.assertEqual(glyph.verticalOrigin, 1000)
        self.assertEqual(glyph.height, 0)
        glyph.verticalOrigin = 0
        self.assertEqual(glyph.verticalOrigin, 0)
        self.assertEqual(glyph.height, 0)
        glyph.verticalOrigin = -10
        self.assertEqual(glyph.verticalOrigin, -10)
        self.assertEqual(glyph.height, 0)
        glyph.verticalOrigin = None
        self.assertIsNone(glyph.verticalOrigin)
        self.assertEqual(glyph.height, 0)

        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertIsNone(glyph.verticalOrigin)
        self.assertEqual(glyph.height, 500)
        glyph.verticalOrigin = 1000
        self.assertEqual(glyph.verticalOrigin, 1000)
        self.assertEqual(glyph.height, 500)
        glyph.verticalOrigin = 0
        self.assertEqual(glyph.verticalOrigin, 0)
        self.assertEqual(glyph.height, 500)
        glyph.verticalOrigin = -10
        self.assertEqual(glyph.verticalOrigin, -10)
        self.assertEqual(glyph.height, 500)
        glyph.verticalOrigin = None
        self.assertIsNone(glyph.verticalOrigin)
        self.assertEqual(glyph.height, 500)

    def test_appendContour(self):
        glyph = Glyph()
        glyph.dirty = False
        contour = Contour()
        glyph.appendContour(contour)
        self.assertEqual(len(glyph), 1)
        self.assertTrue(glyph.dirty)
        self.assertEqual(contour.getParent(), glyph)

    def test_removeContour(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        contour = glyph[0]
        glyph.removeContour(contour)
        self.assertFalse(contour in glyph._contours)
        self.assertIsNone(contour.getParent())

    def test_contourIndex(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        contour = glyph[0]
        self.assertEqual(glyph.contourIndex(contour), 0)
        contour = glyph[1]
        self.assertEqual(glyph.contourIndex(contour), 1)

    def test_clearContours(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.clearContours()
        self.assertEqual(len(glyph), 0)

    def test_components(self):
        font = Font(getTestFontPath())
        glyph = font["C"]
        self.assertEqual(len(glyph.components), 2)

    def test_appendComponent(self):
        glyph = Glyph()
        glyph.dirty = False
        component = Component()
        glyph.appendComponent(component)
        self.assertEqual(len(glyph.components), 1)
        self.assertTrue(glyph.dirty)
        self.assertEqual(component.getParent(), glyph)

    def test_removeComponent(self):
        font = Font(getTestFontPath())
        glyph = font["C"]
        component = glyph.components[0]
        glyph.removeComponent(component)
        self.assertFalse(component in glyph.components)
        self.assertIsNone(component.getParent())

    def test_componentIndex(self):
        font = Font(getTestFontPath())
        glyph = font["C"]
        component = glyph.components[0]
        self.assertEqual(glyph.componentIndex(component), 0)
        component = glyph.components[1]
        self.assertEqual(glyph.componentIndex(component), 1)

    def test_clearComponents(self):
        font = Font(getTestFontPath())
        glyph = font["C"]
        glyph.clearComponents()
        self.assertEqual(len(glyph.components), 0)

    def test_decomposeComponent(self):
        font = Font()

        font.newGlyph("baseGlyph")
        baseGlyph = font["baseGlyph"]
        pointPen = baseGlyph.getPointPen()
        pointPen.beginPath(identifier="contour1")
        pointPen.addPoint((0, 0), "move", identifier="point1")
        pointPen.addPoint((0, 100), "line")
        pointPen.addPoint((100, 100), "line")
        pointPen.addPoint((100, 0), "line")
        pointPen.addPoint((0, 0), "line")
        pointPen.endPath()

        font.newGlyph("referenceGlyph")
        referenceGlyph = font["referenceGlyph"]
        pointPen = referenceGlyph.getPointPen()
        pointPen.addComponent("baseGlyph", (1, 0, 0, 1, 0, 0))
        self.assertEqual(len(referenceGlyph.components), 1)
        self.assertEqual(len(referenceGlyph), 0)
        referenceGlyph.decomposeAllComponents()
        self.assertEqual(len(referenceGlyph.components), 0)
        self.assertEqual(len(referenceGlyph), 1)
        self.assertEqual(referenceGlyph[0].identifier, "contour1")
        self.assertEqual(referenceGlyph[0][0].identifier, "point1")

        pointPen.addComponent("baseGlyph", (1, 0, 0, 1, 100, 100))
        self.assertEqual(len(referenceGlyph.components), 1)
        self.assertEqual(len(referenceGlyph), 1)
        component = referenceGlyph.components[0]
        referenceGlyph.decomposeComponent(component)
        self.assertEqual(len(referenceGlyph.components), 0)
        self.assertEqual(len(referenceGlyph), 2)
        self.assertEqual(referenceGlyph[0].identifier, "contour1")
        self.assertEqual(referenceGlyph[0][0].identifier, "point1")
        referenceGlyph[1].identifier
        referenceGlyph[1][0].identifier

    def test_decomposeComponent_nested_components(self):
        font = Font()
        font.newGlyph("baseGlyph")
        baseGlyph = font["baseGlyph"]
        pointPen = baseGlyph.getPointPen()
        pointPen.beginPath(identifier="contour1")
        pointPen.addPoint((0, 0), "move", identifier="point1")
        pointPen.addPoint((0, 100), "line")
        pointPen.addPoint((100, 100), "line")
        pointPen.addPoint((100, 0), "line")
        pointPen.addPoint((0, 0), "line")
        pointPen.endPath()

        font.newGlyph("referenceGlyph1")
        referenceGlyph1 = font["referenceGlyph1"]
        pointPen = referenceGlyph1.getPointPen()
        pointPen.addComponent("baseGlyph", (1, 0, 0, 1, 3, 6))
        font.newGlyph("referenceGlyph2")
        referenceGlyph2 = font["referenceGlyph2"]
        pointPen = referenceGlyph2.getPointPen()
        pointPen.addComponent("referenceGlyph1", (1, 0, 0, 1, 10, 20))
        referenceGlyph2.decomposeAllComponents()
        self.assertEqual(len(referenceGlyph2.components), 0)
        self.assertEqual(len(referenceGlyph1.components), 1)
        self.assertEqual(len(referenceGlyph2), 1)
        self.assertEqual(referenceGlyph2.bounds, (13, 26, 113, 126))

    def test_anchors(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(len(glyph.anchors), 2)

    def test_appendAnchor(self):
        glyph = Glyph()
        glyph.dirty = False
        anchor = Anchor()
        glyph.appendAnchor(anchor)
        self.assertEqual(len(glyph.anchors), 1)
        self.assertTrue(glyph.dirty)
        self.assertEqual(anchor.getParent(), glyph)

    def test_removeAnchor(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        anchor = glyph.anchors[0]
        glyph.removeAnchor(anchor)
        self.assertFalse(anchor in glyph.anchors)
        self.assertIsNone(anchor.getParent())

    def test_anchorIndex(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        anchor = glyph.anchors[0]
        self.assertEqual(glyph.anchorIndex(anchor), 0)
        anchor = glyph.anchors[1]
        self.assertEqual(glyph.anchorIndex(anchor), 1)

    def test_clearAnchors(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.clearAnchors()
        self.assertEqual(len(glyph.anchors), 0)

    def test_duplicatedAnchors(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        anchor = glyph.anchors[0]
        with self.assertRaises(AssertionError):
            glyph.appendAnchor(anchor)

    def test_appendGuideline(self):
        glyph = Glyph()
        glyph.dirty = False
        guideline = Guideline()
        glyph.appendGuideline(guideline)
        self.assertEqual(len(glyph.guidelines), 1)
        self.assertTrue(glyph.dirty)
        self.assertEqual(guideline.getParent(), glyph)

    def test_removeGuideline(self):
        font = Font(getTestFontPath())
        glyph = font.layers["Layer 1"]["A"]
        guideline = glyph.guidelines[0]
        glyph.removeGuideline(guideline)
        self.assertFalse(guideline in glyph.guidelines)
        self.assertIsNone(guideline.getParent())

    def test_clearGuidelines(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.clearGuidelines()
        self.assertEqual(len(glyph.guidelines), 0)

    def test_duplicatedGuideline(self):
        font = Font(getTestFontPath())
        glyph = font.layers["Layer 1"]["A"]
        guideline = glyph.guidelines[0]
        with self.assertRaises(AssertionError):
            glyph.appendGuideline(guideline)

    def test_len(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(len(glyph), 2)

    def test_iter(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual([len(contour) for contour in glyph], [4, 4])

    def test_copyDataFromGlyph(self):
        source = Glyph()
        source.name = "a"
        source.width = 1
        source.height = 2
        source.unicodes = [3, 4]
        source.note = "test image"
        source.image = dict(fileName="test image", xScale=1, xyScale=1,
                            yxScale=1, yScale=1, xOffset=0, yOffset=0,
                            color=None)
        source.anchors = [dict(x=100, y=200, name="test anchor")]
        source.guidelines = [dict(x=10, y=20, name="test guideline")]
        source.lib = {"foo": "bar"}
        pen = source.getPointPen()
        pen.beginPath()
        pen.addPoint((100, 200), segmentType="line")
        pen.addPoint((300, 400), segmentType="line")
        pen.endPath()
        component = Component()
        component.base = "b"
        source.appendComponent(component)
        dest = Glyph()
        dest.copyDataFromGlyph(source)

        self.assertNotEqual(source.name, dest.name)
        self.assertEqual(source.width, dest.width)
        self.assertEqual(source.height, dest.height)
        self.assertEqual(source.unicodes, dest.unicodes)
        self.assertEqual(source.note, dest.note)
        self.assertEqual(source.image.items(), dest.image.items())
        self.assertEqual([g.items() for g in source.guidelines],
                         [g.items() for g in dest.guidelines])
        self.assertEqual([g.items() for g in source.anchors],
                         [g.items() for g in dest.anchors])
        self.assertEqual(len(source), len(dest))
        self.assertEqual(len(source.components), len(dest.components))
        sourceContours = []
        for contour in source:
            sourceContours.append([])
            for point in contour:
                sourceContours[-1].append((point.x, point.x,
                                           point.segmentType, point.name))
        destContours = []
        for contour in dest:
            destContours.append([])
            for point in contour:
                destContours[-1].append((point.x, point.x,
                                         point.segmentType, point.name))
        self.assertEqual(sourceContours, destContours)
        self.assertEqual(source.components[0].baseGlyph,
                         dest.components[0].baseGlyph)

    def test_clear(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        contour = glyph[0]
        anchor = glyph.anchors[0]
        glyph.clear()
        self.assertEqual(len(glyph), 0)
        self.assertEqual(len(glyph.anchors), 0)
        glyph = font["C"]
        component = glyph.components[0]
        glyph.clear()
        self.assertEqual(len(glyph.components), 0)
        glyph = font.layers["Layer 1"]["A"]
        guideline = glyph.guidelines[0]
        glyph.clear()
        self.assertEqual(len(glyph.guidelines), 0)

        self.assertEqual((contour.getParent(), component.getParent(),
                         anchor.getParent(), guideline.getParent()),
                         (None, None, None, None))
        self.assertEqual((contour.dispatcher, component.dispatcher,
                         anchor.dispatcher, guideline.dispatcher),
                         (None, None, None, None))

    def test_move(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        xMin, yMin, xMax, yMax = glyph.bounds
        glyph.move((100, 50))
        self.assertEqual((xMin+100, yMin+50, xMax+100, yMax+50),
                         glyph.bounds)
        glyph = font["C"]
        xMin, yMin, xMax, yMax = glyph.bounds
        glyph.move((100, 50))
        self.assertEqual((xMin+100, yMin+50, xMax+100, yMax+50), glyph.bounds)

    def test_pointInside(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertTrue(glyph.pointInside((100, 100)))
        self.assertFalse(glyph.pointInside((350, 350)))
        self.assertFalse(glyph.pointInside((-100, -100)))

    def test_area(self):
        font = Font()

        baseGlyph = font.newGlyph("baseGlyph")
        pointPen = baseGlyph.getPointPen()
        pointPen.beginPath()
        pointPen.addPoint((0, 0), "move")
        pointPen.addPoint((0, 100), "line")
        pointPen.addPoint((100, 100), "line")
        pointPen.addPoint((100, 0), "line")
        pointPen.addPoint((0, 0), "line")
        pointPen.endPath()
        self.assertEqual(baseGlyph.area, 10000)

        componentGlyph = font.newGlyph("componentGlyph")
        pointPen = componentGlyph.getPointPen()
        pointPen.addComponent("baseGlyph", [1, 0, 0, 1, 0, 0])
        self.assertEqual(componentGlyph.area, 10000)

    def test_tempLib(self):
        font = Font()

        glyph = font.newGlyph("A")
        glyph.tempLib["foo"] = "bar"
        self.assertEqual(glyph.tempLib, {"foo": "bar"})

        otherGlyph = font.newGlyph("A.other")
        otherGlyph.setDataFromSerialization(glyph.getDataForSerialization())
        self.assertEqual(otherGlyph.tempLib, {"foo": "bar"})


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import
import weakref
from defcon.objects.base import BaseDictObject
from defcon.tools.representations import kerningSide1GroupsRepresentationFactory, kerningSide2GroupsRepresentationFactory, glyphToKerningSide1GroupsRepresentationFactory, glyphToKerningSide2GroupsRepresentationFactory


class Groups(BaseDictObject):

    """
    This object contains all of the groups in a font.

    **This object posts the following notifications:**

    - Groups.Changed
    - Groups.BeginUndo
    - Groups.EndUndo
    - Groups.BeginRedo
    - Groups.EndRedo
    - Groups.GroupSet
    - Groups.GroupDeleted
    - Groups.Cleared
    - Groups.Updated

    This object behaves like a dict. The keys are group names and the
    values are lists of glyph names::

        {
            "myGroup" : ["a", "b"],
            "myOtherGroup" : ["a.alt", "g.alt"],
        }

    The API for interacting with the data is the same as a standard dict.
    For example, to get a list of all group names::

        groupNames = groups.keys()

    To get all groups including the glyph lists::

        for groupName, glyphList in groups.items():

    To get the glyph list for a particular group name::

        glyphList = groups["myGroup"]

    To set the glyph list for a particular group name::

        groups["myGroup"] = ["x", "y", "z"]

    And so on.

    **Note:** You should not modify the group list and expect the object to
    know about it. For example, this could cause your changes to be lost::

        glyphList = groups["myGroups"]
        glyphList.append("n")

    To make sure the change is noticed, reset the list into the object::

        glyphList = groups["myGroups"]
        glyphList.append("n")
        groups["myGroups"] = glyphList

    This may change in the future.
    """

    changeNotificationName = "Groups.Changed"
    beginUndoNotificationName = "Groups.BeginUndo"
    endUndoNotificationName = "Groups.EndUndo"
    beginRedoNotificationName = "Groups.BeginRedo"
    endRedoNotificationName = "Groups.EndRedo"
    setItemNotificationName = "Groups.GroupSet"
    deleteItemNotificationName = "Groups.GroupDeleted"
    clearNotificationName = "Groups.Cleared"
    updateNotificationName = "Groups.Updated"
    representationFactories = {
    "defcon.groups.kerningSide1Groups" : dict(
        factory=kerningSide1GroupsRepresentationFactory,
        destructiveNotifications=("Groups.Changed")
    ),
    "defcon.groups.kerningSide2Groups" : dict(
        factory=kerningSide2GroupsRepresentationFactory,
        destructiveNotifications=("Groups.Changed")
    ),
    "defcon.groups.kerningGlyphToSide1Group" : dict(
        factory=glyphToKerningSide1GroupsRepresentationFactory,
        destructiveNotifications=("Groups.Changed")
    ),
    "defcon.groups.kerningGlyphToSide2Group" : dict(
        factory=glyphToKerningSide2GroupsRepresentationFactory,
        destructiveNotifications=("Groups.Changed")
    ),
}

    def __init__(self, font=None):
        self._font = None
        if font is not None:
            self._font = weakref.ref(font)
        super(Groups, self).__init__()
        self.beginSelfNotificationObservation()

    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        return self.font

    def _get_font(self):
        if self._font is not None:
            return self._font()
        return None

    font = property(_get_font, doc="The :class:`Font` that this object belongs to.")

    # ------------------------
    # Notification Observation
    # ------------------------

    def endSelfNotificationObservation(self):
        super(Groups, self).endSelfNotificationObservation()
        self._font = None


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import unittest
import os
import glob
import tempfile
import shutil
import fs
import fs.copy
import fs.path
from defcon import Font, Glyph, LayerSet, Guideline
from defcon.errors import DefconError
from defcon.tools.notifications import NotificationCenter
from defcon.test.testTools import (
    getTestFontPath, getTestFontCopyPath, makeTestFontCopy,
    openTestFontAsFileSystem, closeTestFontAsFileSystem,
    tearDownTestFontCopy)
from fontTools.ufoLib import UFOReader, UFOWriter, UFOFileStructure
import zipfile
import logging
from fontTools.misc.loggingTools import CapturingLogHandler

try:
    from plistlib import load, dump
except ImportError:
    from plistlib import readPlist as load, writePlist as dump

testFeaturesText = """
@class1 = [a b c d];

feature liga {
    sub f i by fi;
} liga;

@class2 = [x y z];

feature salt {
    sub a by a.alt;
} salt; feature ss01 {sub x by x.alt} ss01;

feature ss02 {sub y by y.alt} ss02;

# feature calt {
#     sub a b' by b.alt;
# } calt;
"""

class FontTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)
        # Python 3 renamed assertRaisesRegexp to assertRaisesRegex,
        # and fires deprecation warnings if a program uses the old name.
        if not hasattr(self, "assertRaisesRegex"):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def tearDown(self):
        if os.path.exists(getTestFontCopyPath()):
            tearDownTestFontCopy()

    def test_set_parent_data_in_glyph(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        self.assertEqual(id(glyph.getParent()), id(font))

    def test_dispatcher(self):
        font = Font()
        self.assertIsInstance(font.dispatcher, NotificationCenter)
        with self.assertRaises(AttributeError):
            self.font.dispatcher = "foo"

    def test_newGlyph(self):
        font = Font(getTestFontPath())
        glyph = font.newGlyph("NewGlyphTest")
        self.assertEqual(glyph.name, "NewGlyphTest")
        self.assertTrue(glyph.dirty)
        self.assertTrue(font.dirty)
        self.assertEqual(sorted(font.keys()), ["A", "B", "C", "NewGlyphTest"])

    def test_insertGlyph(self):
        font = Font(getTestFontPath())
        glyph = Glyph()
        glyph.name = "NewGlyphTest"
        self.assertEqual(sorted(font.keys()), ["A", "B", "C"])
        font.insertGlyph(glyph)
        self.assertEqual(sorted(font.keys()), ["A", "B", "C", "NewGlyphTest"])

    def test_iter(self):
        font = Font(getTestFontPath())
        self.assertEqual(sorted(glyph.name for glyph in font), ["A", "B", "C"])
        names = []
        for glyph1 in font:
            for glyph2 in font:
                names.append((glyph1.name, glyph2.name))
        self.assertEqual(sorted(names),
                         [("A", "A"), ("A", "B"), ("A", "C"),
                          ("B", "A"), ("B", "B"), ("B", "C"),
                          ("C", "A"), ("C", "B"), ("C", "C")])

    def test_getitem(self):
        font = Font(getTestFontPath())
        self.assertEqual(font["A"].name, "A")
        self.assertEqual(font["B"].name, "B")
        with self.assertRaises(KeyError):
            font["NotInFont"]

    def test_delitem(self):
        path = makeTestFontCopy()
        font = Font(path)
        del font["A"]
        self.assertTrue(font.dirty)
        font.newGlyph("NewGlyphTest")
        del font["NewGlyphTest"]
        self.assertEqual(sorted(font.keys()), ["B", "C"])
        self.assertEqual(len(font), 2)
        self.assertFalse("A" in font)
        font.save()
        fileNames = glob.glob(os.path.join(path, 'glyphs', '*.glif'))
        fileNames = [os.path.basename(fileName) for fileName in fileNames]
        self.assertEqual(sorted(fileNames), ["B_.glif", "C_.glif"])
        with self.assertRaises(KeyError):
            del font["NotInFont"]

    def test_delitem_glyph_not_dirty(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                font["A"]  # glyph = font["A"]
                fileSystem = openTestFontAsFileSystem(path)
                glyphPath = fs.path.join("glyphs", "A_.glif")
                fileSystem.remove(glyphPath)
                contentsPath = fs.path.join("glyphs", "contents.plist")
                with fileSystem.open(contentsPath, "rb") as f:
                    plist = load(f)
                del plist["A"]
                with fileSystem.open(contentsPath, "wb") as f:
                    dump(plist, f)
                closeTestFontAsFileSystem(fileSystem, path)
                r = font.testForExternalChanges()
                self.assertEqual(r["deletedGlyphs"], ["A"])
                del font["A"]
                font.save()
                self.assertFalse(os.path.exists(glyphPath))
            tearDownTestFontCopy(font.path)

    def test_delitem_glyph_dirty(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                glyph = font["A"]
                glyph.dirty = True
                fileSystem = openTestFontAsFileSystem(path)
                glyphPath = fs.path.join("glyphs", "A_.glif")
                fileSystem.remove(glyphPath)
                contentsPath = fs.path.join("glyphs", "contents.plist")
                with fileSystem.open(contentsPath, "rb") as f:
                    plist = load(f)
                del plist["A"]
                with fileSystem.open(contentsPath, "wb") as f:
                    dump(plist, f)
                closeTestFontAsFileSystem(fileSystem, path)
                r = font.testForExternalChanges()
                self.assertEqual(r["deletedGlyphs"], ["A"])
                del font["A"]
                font.save()
                fileSystem = openTestFontAsFileSystem(path)
                self.assertFalse(fileSystem.exists(glyphPath))
                closeTestFontAsFileSystem(fileSystem, path)
            tearDownTestFontCopy(font.path)

    def test_len(self):
        font = Font(getTestFontPath())
        self.assertEqual(len(font), 3)
        font = Font()
        self.assertEqual(len(font), 0)

    def test_contains(self):
        font = Font(getTestFontPath())
        self.assertTrue("A" in font)
        self.assertFalse("NotInFont" in font)
        font = Font()
        self.assertFalse("A" in font)

    def test_keys(self):
        font = Font(getTestFontPath())
        self.assertEqual(sorted(font.keys()), ["A", "B", "C"])
        del font["A"]
        self.assertEqual(sorted(font.keys()), ["B", "C"])
        font.newGlyph("A")
        self.assertEqual(sorted(font.keys()), ["A", "B", "C"])
        font = Font()
        self.assertEqual(font.keys(), set())
        font.newGlyph("A")
        self.assertEqual(sorted(font.keys()), ["A"])

    def test_path_get(self):
        path = getTestFontPath()
        font = Font(path)
        self.assertEqual(font.path, path)
        font = Font()
        self.assertIsNone(font.path)

    def test_path_set(self):
        path1 = getTestFontPath()
        font = Font(path1)
        path2 = getTestFontPath("setPathTest.ufo")
        shutil.copytree(path1, path2)
        font.path = path2
        self.assertEqual(font.path, path2)
        shutil.rmtree(path2)

    def test_glyphsWithOutlines(self):
        font = Font(getTestFontPath())
        self.assertEqual(sorted(font.glyphsWithOutlines), ["A", "B"])
        font = Font(getTestFontPath())
        for glyph in font:
            pass
        self.assertEqual(sorted(font.glyphsWithOutlines), ["A", "B"])

    def test_componentReferences(self):
        font = Font(getTestFontPath())
        self.assertEqual(font.componentReferences,
                         {"A": set(["C"]), "B": set(["C"])})

    def test_bounds(self):
        font = Font(getTestFontPath())
        self.assertEqual(font.bounds, (0, 0, 700, 700))

    def test_controPointsBounds(self):
        font = Font(getTestFontPath())
        self.assertEqual(font.controlPointBounds, (0, 0, 700, 700))

    def test_beginSelfLayerSetNotificationObservation(self):
        font = Font()
        self.assertTrue(font.dispatcher.hasObserver(
            font, "LayerSet.Changed", font.layers))
        self.assertTrue(font.dispatcher.hasObserver(
            font, "LayerSet.LayerAdded", font.layers))
        self.assertTrue(font.dispatcher.hasObserver(
            font, "LayerSet.LayerWillBeDeleted", font.layers))

        font.layers.removeObserver(
            observer=self, notification="LayerSet.Changed")
        font.layers.removeObserver(
            observer=self, notification="LayerSet.LayerAdded")
        font.layers.removeObserver(
            observer=self, notification="LayerSet.LayerWillBeDeleted")
        font.layers.endSelfNotificationObservation()

        font.beginSelfLayerSetNotificationObservation()
        self.assertTrue(font.dispatcher.hasObserver(
            font, "LayerSet.Changed", font.layers))
        self.assertTrue(font.dispatcher.hasObserver(
            font, "LayerSet.LayerAdded", font.layers))
        self.assertTrue(font.dispatcher.hasObserver(
            font, "LayerSet.LayerWillBeDeleted", font.layers))

    def test_endSelfLayerSetNotificationObservation(self):
        font = Font()
        font.endSelfLayerSetNotificationObservation()
        self.assertFalse(font.dispatcher.hasObserver(
            font, "LayerSet.Changed", font.layers))
        self.assertFalse(font.dispatcher.hasObserver(
            font, "LayerSet.LayerAdded", font.layers))
        self.assertFalse(font.dispatcher.hasObserver(
            font, "LayerSet.LayerWillBeDeleted", font.layers))

    def test_layers(self):
        font = Font(getTestFontPath())
        self.assertIsInstance(font.layers, LayerSet)
        self.assertEqual(font.layers.layerOrder,
                         ["public.default", "public.background", "Layer 1"])
        self.assertTrue(font.layers.hasObserver(font, "LayerSet.Changed"))
        self.assertTrue(font.layers.hasObserver(font, "LayerSet.LayerAdded"))
        self.assertTrue(font.layers.hasObserver(font,
                                                "LayerSet.LayerWillBeDeleted"))

    def test_font_observes_new_layer(self):
        font = Font()
        font.layers.newLayer("test_layer")
        layer = font.layers["test_layer"]
        self.assertTrue(layer.hasObserver(font, "Layer.GlyphAdded"))

    def test_font_observes_loaded_layers(self):
        font = Font(getTestFontPath())
        for layername in font.layers.layerOrder:
            layer = font.layers[layername]
            self.assertTrue(layer.hasObserver(font, "Layer.GlyphAdded"))

    def test_glyphOrder(self):
        font = Font(getTestFontPath())
        self.assertEqual(font.glyphOrder, [])
        font.glyphOrder = sorted(font.keys())
        self.assertEqual(font.glyphOrder, ["A", "B", "C"])
        layer = font.layers.defaultLayer
        layer.newGlyph("X")
        self.assertEqual(sorted(layer.keys()), ["A", "B", "C", "X"])
        self.assertEqual(font.glyphOrder, ["A", "B", "C", "X"])
        del layer["A"]
        self.assertEqual(font.glyphOrder, ["A", "B", "C", "X"])
        del layer["X"]
        self.assertEqual(font.glyphOrder, ["A", "B", "C"])
        layer["B"].name = "Y"
        self.assertEqual(font.glyphOrder, ["A", "Y", "C"])

    def test_updateGlyphOrder_none(self):
        font = Font(getTestFontPath())
        self.assertEqual(font.glyphOrder, [])
        font.updateGlyphOrder()
        self.assertEqual(font.glyphOrder, [])

    def test_updateGlyphOrder_add(self):
        font = Font(getTestFontPath())
        self.assertEqual(font.glyphOrder, [])
        font.updateGlyphOrder(addedGlyph="test")
        self.assertEqual(font.glyphOrder, ["test"])

    def test_updateGlyphOrder_remove(self):
        font = Font(getTestFontPath())
        self.assertEqual(font.glyphOrder, [])
        font.glyphOrder = ["test"]
        self.assertEqual(font.glyphOrder, ["test"])
        font.updateGlyphOrder(removedGlyph="test")
        self.assertEqual(font.glyphOrder, [])

    def test_updateGlyphOrder_rename(self):
        font = Font(getTestFontPath())
        self.assertEqual(font.glyphOrder, [])
        font.glyphOrder = sorted(font.keys())
        self.assertEqual(font.glyphOrder, ["A", "B", "C"])
        font.updateGlyphOrder(addedGlyph="new", removedGlyph="B")
        self.assertEqual(font.glyphOrder, ["A", "new", "C"])

    def test_guidelines(self):
        font = Font(getTestFontPath())
        self.assertEqual(font.guidelines, [])
        guideline1 = Guideline(guidelineDict={"x": 100})
        guideline2 = Guideline(guidelineDict={"y": 200})
        font.guidelines = [guideline1, guideline2]
        self.assertEqual(font.guidelines, [guideline1, guideline2])

    def test_instantiateGuideline(self):
        font = Font(getTestFontPath())
        guideline = font.instantiateGuideline()
        self.assertIsInstance(guideline, Guideline)
        guideline = font.instantiateGuideline(guidelineDict={"x": 100})
        self.assertEqual(dict(guideline), {'x': 100})

    def test_beginSelfGuidelineNotificationObservation(self):
        font = Font(getTestFontPath())
        guideline = font.instantiateGuideline()
        self.assertFalse(guideline.dispatcher.hasObserver(
            font, "Guideline.Changed", guideline))
        font.beginSelfGuidelineNotificationObservation(guideline)
        self.assertTrue(guideline.dispatcher.hasObserver(
            font, "Guideline.Changed", guideline))

    def test_endSelfGuidelineNotificationObservation(self):
        font = Font(getTestFontPath())
        guideline = font.instantiateGuideline()
        font.beginSelfGuidelineNotificationObservation(guideline)
        self.assertTrue(guideline.hasObserver(
            font, "Guideline.Changed"))
        font.endSelfGuidelineNotificationObservation(guideline)
        self.assertIsNone(guideline.dispatcher)
        self.assertFalse(guideline.hasObserver(
            font, "Guideline.Changed"))

    def test_appendGuideline(self):
        font = Font(getTestFontPath())
        guideline1 = Guideline(guidelineDict={"x": 100})
        font.appendGuideline(guideline1)
        self.assertEqual([dict(guideline) for guideline in font.guidelines], [{'x': 100}])
        guideline2 = Guideline(guidelineDict={"y": 200})
        font.appendGuideline(guideline2)
        self.assertEqual([dict(guideline) for guideline in font.guidelines], [{'x': 100}, {'y': 200}])
        guideline3 = Guideline(guidelineDict={"y": 100})
        font.appendGuideline(guideline3)
        self.assertEqual([dict(guideline) for guideline in font.guidelines], [{'x': 100}, {'y': 200}, {'y': 100}])

    def test_insertGuideline(self):
        font = Font(getTestFontPath())
        guideline1 = Guideline(guidelineDict={"x": 100})
        font.insertGuideline(0, guideline1)
        self.assertEqual([dict(guideline) for guideline in font.guidelines], [{'x': 100}])
        guideline2 = Guideline(guidelineDict={"y": 200})
        font.insertGuideline(0, guideline2)
        self.assertEqual([dict(guideline) for guideline in font.guidelines], [{'y': 200}, {'x': 100}])
        guideline3 = Guideline(guidelineDict={"y": 100})
        font.insertGuideline(2, guideline3)
        self.assertEqual([dict(guideline) for guideline in font.guidelines], [{'y': 200}, {'x': 100}, {'y': 100}])

    def test_removeGuideline(self):
        font = Font(getTestFontPath())
        guideline1 = Guideline(guidelineDict={"x": 100})
        guideline2 = Guideline(guidelineDict={"y": 200})
        font.guidelines = [guideline1, guideline2]
        font.removeGuideline(guideline1)
        self.assertEqual(font.guidelines, [guideline2])

    def test_guidelineIndex(self):
        font = Font(getTestFontPath())
        guideline1 = Guideline(guidelineDict={"x": 100})
        guideline2 = Guideline(guidelineDict={"y": 200})
        font.guidelines = [guideline1, guideline2]
        self.assertEqual(font.guidelineIndex(guideline1), 0)
        self.assertEqual(font.guidelineIndex(guideline2), 1)

    def test_clearGuidelines(self):
        font = Font(getTestFontPath())
        guideline1 = Guideline(guidelineDict={"x": 100})
        guideline2 = Guideline(guidelineDict={"y": 200})
        font.guidelines = [guideline1, guideline2]
        self.assertEqual(font.guidelines, [guideline1, guideline2])
        font.clearGuidelines()
        self.assertEqual(font.guidelines, [])

    def test_save(self):
        for ufo in (u"TestFont.ufo", u"TestFont.ufoz"):
            path = makeTestFontCopy(getTestFontPath(ufo))
            try:
                with Font(path) as font:
                    origFileStructure = font.ufoFileStructure
                    for glyph in font:
                        glyph.dirty = True
                    font.save()
                    fileNames = sorted(
                        [
                            fs.path.basename(m.path)
                            for m in UFOReader(path).fs.glob("glyphs/*.glif")
                        ]
                    )
                    self.assertEqual(fileNames, ["A_.glif", "B_.glif", "C_.glif"])
                    self.assertEqual(origFileStructure, font.ufoFileStructure)
            finally:
                tearDownTestFontCopy(path)

    def test_save_as(self):
        for ufo in (u"TestFont.ufo", u"TestFont.ufoz"):
            path = getTestFontPath(ufo)
            font = Font(path)
            origFileStructure = font.ufoFileStructure
            saveAsPath = getTestFontCopyPath(path)
            self.assertFalse(os.path.exists(saveAsPath))
            font.save(saveAsPath)
            try:
                fileNames = sorted(
                    [
                        fs.path.basename(m.path)
                        for m in UFOReader(saveAsPath).fs.glob("glyphs/*.glif")
                    ]
                )
                self.assertEqual(fileNames, ["A_.glif", "B_.glif", "C_.glif"])
                self.assertEqual(font.path, saveAsPath)
                self.assertEqual(origFileStructure, font.ufoFileStructure)
            finally:
                font.close()
                tearDownTestFontCopy(saveAsPath)

    def test_save_same_path(self):
        for ufo in (u"TestFont.ufo", u"TestFont.ufoz"):
            path = makeTestFontCopy(getTestFontPath(ufo))
            isZip = zipfile.is_zipfile(path)
            font = Font(path)
            try:
                font = Font(path)
                font.save(path)
                if isZip:
                    self.assertTrue(zipfile.is_zipfile(path))
                else:
                    self.assertTrue(os.path.isdir(path))
            finally:
                font.close()
                tearDownTestFontCopy(path)

    def test_save_same_path_different_structure(self):
        for ufo in ("TestFont.ufo", "TestFont.ufoz"):
            path = makeTestFontCopy(getTestFontPath(ufo))
            isZip = zipfile.is_zipfile(path)
            with Font(path) as font:
                with self.assertRaisesRegex(
                    DefconError,
                    "Can't save font in-place with a different structure"
                ):
                    font.save(path, structure="package" if isZip else "zip")
            tearDownTestFontCopy(path)

    def test_save_new_font_without_path(self):
        font = Font()
        msg = "Can't save new font without a 'path'"
        with self.assertRaisesRegex(DefconError, msg):
            font.save()

    def test_save_new_font_to_exsisting_directory(self):
        for ufo in ("TestFont.ufo", "TestFont.ufoz"):
            path = makeTestFontCopy(getTestFontPath(ufo))
            font = Font()
            try:
                self.assertTrue(os.path.exists(path))
                font.save(path)
                self.assertTrue(os.path.isdir(path))
            finally:
                font.close()
                tearDownTestFontCopy(path)

    def test_save_ufoz(self):
        path = getTestFontPath()
        tmpdir = tempfile.mkdtemp()
        dest = os.path.join(tmpdir, "TestFont.ufoz")
        font = Font(path)
        try:
            self.assertFalse(os.path.exists(dest))
            self.assertEqual(font.path, path)
            font.save(dest, structure="zip")
            self.assertTrue(os.path.exists(dest))
            self.assertTrue(zipfile.is_zipfile(dest))
            self.assertEqual(font.path, dest)
            self.assertEqual(font.ufoFileStructure, UFOFileStructure.ZIP)
            fileNames = sorted(
                [
                    fs.path.basename(m.path)
                    for m in UFOReader(dest).fs.glob("glyphs/*.glif")
                ]
            )
            self.assertEqual(fileNames, ["A_.glif", "B_.glif", "C_.glif"])
        finally:
            font.close()
            shutil.rmtree(tmpdir)

    def test_save_new_font_to_existing_file(self):
        with tempfile.NamedTemporaryFile(delete=False) as tmp:
            pass
        path = tmp.name
        self.assertTrue(os.path.exists(path))
        try:
            font = Font()
            font.save(path)
            self.assertTrue(os.path.isdir(path))
        finally:
            if os.path.isfile(path):
                os.remove(path)
            elif os.path.isdir(path):
                shutil.rmtree(path)

    def test_new_font_format(self):
        font = Font()
        self.assertEqual(font.ufoFormatVersion, None)
        self.assertEqual(font.ufoFormatVersionTuple, None)

    def test_save_in_place_different_format(self):
        path = makeTestFontCopy()
        font = Font(path)
        self.assertEqual(font.ufoFormatVersion, 3)
        self.assertEqual(font.ufoFormatVersionTuple, (3, 0))
        font.save(formatVersion=2)
        self.assertEqual(font.ufoFormatVersion, 2)
        self.assertEqual(font.ufoFormatVersionTuple, (2, 0))

    def test_save_in_place_invalid_ufo(self):
        path = makeTestFontCopy()
        font = Font(path)
        layercontents = os.path.join(path, "layercontents.plist")
        os.remove(layercontents)
        self.assertFalse(os.path.exists(layercontents))

        logger = logging.getLogger("defcon.objects.font")
        with CapturingLogHandler(logger, level="ERROR") as captor:
            font.save()
        captor.assertRegex("Invalid ufo found")

        self.assertTrue(os.path.exists(layercontents))
        font = Font(path)
        _ = font.layers
        font.save()

    def test_save_ufo3z_as_ufo2(self):
        # https://github.com/robotools/defcon/issues/296
        font = Font()
        font.layers.defaultLayer.name = "a-custom-layer-name"
        font.newLayer("a-new-layer")

        with tempfile.TemporaryDirectory() as root:
            path_ufo3 = os.path.join(root, "three.ufo")
            font.save(path_ufo3, formatVersion=3, structure="zip")
            font.close()

            path_ufo2 = os.path.join(root, "two.ufo")

            with Font(path_ufo3) as font_ufo2:
                font_ufo2.save(path_ufo2, formatVersion=2)


    def test_testForExternalChanges(self):
        for ufo in (u"TestExternalEditing.ufo", u"TestExternalEditing.ufoz"):
            path = getTestFontPath(ufo)
            path = makeTestFontCopy(path)
            with Font(path) as font:
                # load all the objects so that they get stamped
                font.info  # i = font.info
                k = font.kerning
                font.groups  # g = font.groups
                font.lib  # l = font.lib
                font["A"]  # g = font["A"]

                d = font.testForExternalChanges()
                self.assertFalse(d["info"])
                self.assertFalse(d["kerning"])
                self.assertFalse(d["groups"])
                self.assertFalse(d["lib"])

                # make a simple change to the kerning data
                fileSystem = openTestFontAsFileSystem(font.path)
                path = u"kerning.plist"
                t = fileSystem.readbytes(path)
                t += b"<!-- test -->"
                fileSystem.writebytes(path, t)
                k._dataOnDiskTimeStamp -= 1
                closeTestFontAsFileSystem(fileSystem, font.path)

                d = font.testForExternalChanges()
                self.assertTrue(d["kerning"])
                self.assertFalse(d["groups"])
                self.assertFalse(d["info"])
                self.assertFalse(d["lib"])

                # save the kerning data and test again
                font.kerning.dirty = True
                font.save()
                d = font.testForExternalChanges()
                self.assertFalse(d["kerning"])
                self.assertFalse(d["groups"])
                self.assertFalse(d["info"])
                self.assertFalse(d["lib"])

            tearDownTestFontCopy(font.path)

    def test_reloadInfo(self):
        path = getTestFontPath(u"TestExternalEditing.ufo")
        font = Font(path)
        info = font.info

        path = os.path.join(font.path, "fontinfo.plist")
        f = open(path, "r")
        t = f.read()
        f.close()
        t = t.replace("<integer>750</integer>", "<integer>751</integer>")
        f = open(path, "w")
        f.write(t)
        f.close()

        self.assertEqual(info.ascender, 750)
        font.reloadInfo()
        self.assertEqual(info.ascender, 751)

        t = t.replace("<integer>751</integer>", "<integer>750</integer>")
        f = open(path, "w")
        f.write(t)
        f.close()

    def test_reloadKerning(self):
        path = getTestFontPath(u"TestExternalEditing.ufo")
        font = Font(path)
        kerning = font.kerning

        path = os.path.join(font.path, "kerning.plist")
        f = open(path, "r")
        t = f.read()
        f.close()
        t = t.replace("<integer>-100</integer>", "<integer>-101</integer>")
        f = open(path, "w")
        f.write(t)
        f.close()

        self.assertEqual(list(kerning.items()), [(("A", "A"), -100)])
        font.reloadKerning()
        self.assertEqual(list(kerning.items()), [(("A", "A"), -101)])

        t = t.replace("<integer>-101</integer>", "<integer>-100</integer>")
        f = open(path, "w")
        f.write(t)
        f.close()

    def test_reloadGroups(self):
        path = getTestFontPath(u"TestExternalEditing.ufo")
        font = Font(path)
        groups = font.groups

        path = os.path.join(font.path, "groups.plist")
        f = open(path, "r")
        t = f.read()
        f.close()
        t = t.replace("<key>TestGroup</key>", "<key>XXX</key>")
        f = open(path, "w")
        f.write(t)
        f.close()

        self.assertEqual(list(groups.keys()), ["TestGroup"])
        font.reloadGroups()
        self.assertEqual(list(groups.keys()), ["XXX"])

        t = t.replace("<key>XXX</key>", "<key>TestGroup</key>")
        f = open(path, "w")
        f.write(t)
        f.close()

    def test_reloadLib(self):
        path = getTestFontPath(u"TestExternalEditing.ufo")
        font = Font(path)
        lib = font.lib

        path = os.path.join(font.path, "lib.plist")
        f = open(path, "r")
        t = f.read()
        f.close()
        t = t.replace("<key>org.robofab.glyphOrder</key>",
                      "<key>org.robofab.glyphOrder.XXX</key>")
        f = open(path, "w")
        f.write(t)
        f.close()

        self.assertEqual(list(lib.keys()), ["org.robofab.glyphOrder"])
        font.reloadLib()
        self.assertEqual(list(lib.keys()), ["org.robofab.glyphOrder.XXX"])

        t = t.replace("<key>org.robofab.glyphOrder.XXX</key>",
                      "<key>org.robofab.glyphOrder</key>")
        f = open(path, "w")
        f.write(t)
        f.close()

    def test_reloadGlyphs(self):
        path = getTestFontPath(u"TestExternalEditing.ufo")
        font = Font(path)
        glyph = font["A"]

        path = os.path.join(font.path, "glyphs", "A_.glif")
        f = open(path, "r")
        t = f.read()
        f.close()
        t = t.replace('<advance width="700"/>', '<advance width="701"/>')
        f = open(path, "w")
        f.write(t)
        f.close()

        self.assertEqual(glyph.width, 700)
        self.assertEqual(len(glyph), 2)
        font.reloadGlyphs(["A"])
        self.assertEqual(glyph.width, 701)
        self.assertEqual(len(glyph), 2)

        t = t.replace('<advance width="701"/>', '<advance width="700"/>')
        f = open(path, "w")
        f.write(t)
        f.close()

    def test_splitFeaturesForConversion(self):
        font = Font()
        self.assertEqual(
            font._splitFeaturesForConversion(testFeaturesText),
            (
                "\n@class1 = [a b c d];\n",
                [("liga", "\nfeature liga {\n    sub f i by fi;\n} liga;\n\n"
                  "@class2 = [x y z];\n"),
                 ("salt", "\nfeature salt {\n    sub a by a.alt;\n} salt; "
                  "feature ss01 {sub x by x.alt} ss01;\n"),
                 ("ss02", "\nfeature ss02 {sub y by y.alt} ss02;\n\n"
                  "# feature calt {\n#     sub a b' by b.alt;\n# } calt;\n")]
            )
        )

    def test_glyph_name_change(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.name = "NameChangeTest"
        self.assertEqual(sorted(font.keys()), ["B", "C", "NameChangeTest"])
        self.assertTrue(font.dirty)

    def test_glyph_unicodes_changed(self):
        font = Font(getTestFontPath())
        glyph = font["A"]
        glyph.unicodes = [123, 456]
        self.assertEqual(font.unicodeData[123], ["A"])
        self.assertEqual(font.unicodeData[456], ["A"])
        self.assertEqual(font.unicodeData[66], ["B"])

        font = Font(getTestFontPath())
        glyph = font.newGlyph("test")
        glyph.unicodes = [65]
        self.assertEqual(font.unicodeData[65], ["test", "A"])

    def test_tempLib(self):
        font = Font()

        font.tempLib["foo"] = "bar"
        self.assertEqual(font.tempLib, {"foo": "bar"})

        otherFont = Font()
        otherFont.setDataFromSerialization(font.getDataForSerialization())
        self.assertEqual(otherFont.tempLib, {"foo": "bar"})


if __name__ == "__main__":
    unittest.main()



from __future__ import absolute_import
import weakref
from warnings import warn
from fontTools.misc.arrayTools import unionRect
from defcon.objects.base import BaseObject
from defcon.objects.contour import Contour
from defcon.objects.point import Point
from defcon.objects.component import Component
from defcon.objects.anchor import Anchor
from defcon.objects.lib import Lib
from defcon.objects.guideline import Guideline
from defcon.objects.image import Image
from defcon.objects.color import Color
from defcon.tools.representations import glyphAreaRepresentationFactory
from defcon.pens.decomposeComponentPointPen import DecomposeComponentPointPen


def addRepresentationFactory(name, factory):
    warn("addRepresentationFactory is deprecated. Use the functions in defcon.__init__.", DeprecationWarning)
    Glyph.representationFactories[name] = dict(factory=factory, destructiveNotifications=["Glyph.Changed"])

def removeRepresentationFactory(name):
    warn("removeRepresentationFactory is deprecated. Use the functions in defcon.__init__.", DeprecationWarning)
    del Glyph.representationFactories[name]


class Glyph(BaseObject):

    """
    This object represents a glyph and it contains contour, component, anchor
    and other assorted bits data about the glyph.

    **This object posts the following notifications:**

    - Glyph.Changed
    - Glyph.BeginUndo
    - Glyph.EndUndo
    - Glyph.BeginRedo
    - Glyph.EndRedo
    - Glyph.NameWillChange
    - Glyph.NameChanged
    - Glyph.UnicodesChanged
    - Glyph.WidthChanged
    - Glyph.HeightChanged
    - Glyph.LeftMarginWillChange
    - Glyph.LeftMarginDidChange
    - Glyph.RightMarginWillChange
    - Glyph.RightMarginDidChange
    - Glyph.TopMarginWillChange
    - Glyph.TopMarginDidChange
    - Glyph.BottomMarginWillChange
    - Glyph.BottomMarginDidChange
    - Glyph.NoteChanged
    - Glyph.LibChanged
    - Glyph.ImageChanged
    - Glyph.ImageWillBeCleared
    - Glyph.ImageCleared
    - Glyph.ContourWillBeAdded
    - Glyph.ContourWillBeDeleted
    - Glyph.ContoursChanged
    - Glyph.ComponentWillBeAdded
    - Glyph.ComponentWillBeDeleted
    - Glyph.ComponentsChanged
    - Glyph.AnchorWillBeAdded
    - Glyph.AnchorWillBeDeleted
    - Glyph.AnchorsChanged
    - Glyph.GuidelineWillBeAdded
    - Glyph.GuidelineWillBeDeleted
    - Glyph.GuidelinesChanged
    - Glyph.MarkColorChanged
    - Glyph.VerticalOriginChanged

    The Glyph object has list like behavior. This behavior allows you to interact
    with contour data directly. For example, to get a particular contour::

        contour = glyph[0]

    To iterate over all contours::

        for contour in glyph:

    To get the number of contours::

        contourCount = len(glyph)

    To interact with components or anchors in a similar way,
    use the ``components`` and ``anchors`` attributes.
    """

    changeNotificationName = "Glyph.Changed"
    beginUndoNotificationName = "Glyph.BeginUndo"
    endUndoNotificationName = "Glyph.EndUndo"
    beginRedoNotificationName = "Glyph.BeginRedo"
    endRedoNotificationName = "Glyph.EndRedo"
    representationFactories = {
        "defcon.glyph.area" : dict(
            factory=glyphAreaRepresentationFactory,
            destructiveNotifications=("Glyph.ContoursChanged", "Glyph.ComponentsChanged", "Glyph.ComponentBaseGlyphDataChanged")
        )
    }

    def __init__(self, layer=None,
        contourClass=None, pointClass=None, componentClass=None, anchorClass=None,
        guidelineClass=None, libClass=None, imageClass=None):
        layerSet = font = None
        if layer is not None:
            layerSet = layer.layerSet
            if layerSet is not None:
                font = weakref.ref(layer.layerSet.font)
                layerSet = weakref.ref(layer.layerSet)
            layer = weakref.ref(layer)
        self._font = font
        self._layerSet = layerSet
        self._layer = layer
        super(Glyph, self).__init__()
        self.beginSelfNotificationObservation()

        self._isLoading = False
        self._dirty = False
        self._name = None
        self._unicodes = []
        self._width = 0
        self._height = 0
        self._note = None
        self._image = None
        self._identifiers = set()
        self._shallowLoadedContours = None
        self._contours = []
        self._components = []
        self._anchors = []
        self._guidelines = []
        self._lib = None
        self._tempLib = None

        if contourClass is None:
            contourClass = Contour
        if pointClass is None:
            pointClass = Point
        if componentClass is None:
            componentClass = Component
        if anchorClass is None:
            anchorClass = Anchor
        if guidelineClass is None:
            guidelineClass = Guideline
        if libClass is None:
            libClass = Lib
        if imageClass is None:
            imageClass = Image
        self._contourClass = contourClass
        self._pointClass = pointClass
        self._componentClass = componentClass
        self._anchorClass = anchorClass
        self._guidelineClass = guidelineClass
        self._libClass = libClass
        self._imageClass = imageClass

    def __del__(self):
        super(Glyph, self).__del__()
        self._contours = None
        self._components = None
        self._anchors = None
        self._guidelines = None
        self._lib = None
        self._image = None

    # --------------
    # Parent Objects
    # --------------

    def getParent(self):
        return self.font

    def _get_font(self):
        if self._font is None:
            return None
        return self._font()

    font = property(_get_font, doc="The :class:`Font` that this glyph belongs to.")

    def _get_layerSet(self):
        if self._layerSet is None:
            return None
        return self._layerSet()

    layerSet = property(_get_layerSet, doc="The :class:`LayerSet` that this glyph belongs to.")

    def _get_layer(self):
        if self._layer is None:
            return None
        return self._layer()

    layer = property(_get_layer, doc="The :class:`Layer` that this glyph belongs to.")

    # ----------------
    # Basic Attributes
    # ----------------

    # identifiers

    def _get_identifiers(self):
        return self._identifiers

    identifiers = property(_get_identifiers, doc="Set of identifiers for the glyph. This is primarily for internal use.")

    # name

    def _set_name(self, value):
        oldName = self._name
        if oldName != value:
            self.postNotification(notification="Glyph.NameWillChange", data=dict(oldValue=oldName, newValue=value))
            self._name = value
            self.postNotification(notification="Glyph.NameChanged", data=dict(oldValue=oldName, newValue=value))
            self.dirty = True

    def _get_name(self):
        return self._name

    name = property(_get_name, _set_name, doc="The name of the glyph. Setting this posts *GLyph.NameChanged* and *Glyph.NameChanged* notifications.")

    # unicodes

    def _get_unicodes(self):
        return list(self._unicodes)

    def _set_unicodes(self, value):
        oldValue = self.unicodes
        if oldValue != value:
            self._unicodes = list(value)
            self.postNotification(notification="Glyph.UnicodesChanged", data=dict(oldValue=oldValue, newValue=value))
            self.dirty = True

    unicodes = property(_get_unicodes, _set_unicodes, doc="The list of unicode values assigned to the glyph. Setting this posts *Glyph.UnicodesChanged* and *Glyph.Changed* notifications.")

    def _get_unicode(self):
        if self._unicodes:
            return self._unicodes[0]
        return None

    def _set_unicode(self, value):
        if value is None:
            self.unicodes = []
        else:
            self.unicodes = [value]

    unicode = property(_get_unicode, _set_unicode, doc="The primary unicode value for the glyph. This is the equivalent of ``glyph.unicodes[0]``. This is a convenience attribute that works with the ``unicodes`` attribute.")

    # -------
    # Metrics
    # -------

    # bounds

    def _getContourComponentBounds(self, attr):
        bounds = None
        subObjects = [contour for contour in self]
        subObjects += [component for component in self.components]
        for subObject in subObjects:
            b = getattr(subObject, attr)
            if b is not None:
                if bounds is None:
                    bounds = b
                else:
                    bounds = unionRect(bounds, b)
        return bounds

    def _get_bounds(self):
        return self._getContourComponentBounds("bounds")

    bounds = property(_get_bounds, doc="The bounds of the glyph's outline expressed as a tuple of form (xMin, yMin, xMax, yMax).")

    def _get_controlPointBounds(self):
        return self._getContourComponentBounds("controlPointBounds")

    controlPointBounds = property(_get_controlPointBounds, doc="The control bounds of all points in the glyph. This only measures the point positions, it does not measure curves. So, curves without points at the extrema will not be properly measured.")

    # area

    def _get_area(self):
        return self.getRepresentation("defcon.glyph.area")

    area = property(_get_area, doc="The area of the glyph's outline.")

    # margins

    def _get_leftMargin(self):
        bounds = self.bounds
        if bounds is None:
            return None
        xMin, yMin, xMax, yMax = bounds
        return xMin

    def _set_leftMargin(self, value):
        bounds = self.bounds
        if bounds is None:
            return
        xMin, yMin, xMax, yMax = bounds
        oldValue = xMin
        diff = value - xMin
        if value != oldValue:
            self.postNotification(notification="Glyph.LeftMarginWillChange", data=dict(oldValue=oldValue, newValue=value))
            self.move((diff, 0))
            self.width += diff
            self.dirty = True
            self.postNotification(notification="Glyph.LeftMarginDidChange", data=dict(oldValue=oldValue, newValue=value))

    leftMargin = property(_get_leftMargin, _set_leftMargin, doc="The left margin of the glyph. Setting this posts *Glyph.WidthChanged*, *Glyph.LeftMarginWillChange*, *Glyph.LeftMarginDidChange* and *Glyph.Changed* notifications among others.")

    def _get_rightMargin(self):
        bounds = self.bounds
        if bounds is None:
            return None
        xMin, yMin, xMax, yMax = bounds
        return self._width - xMax

    def _set_rightMargin(self, value):
        bounds = self.bounds
        if bounds is None:
            return
        xMin, yMin, xMax, yMax = bounds
        oldValue = self._width - xMax
        if oldValue != value:
            self.postNotification(notification="Glyph.RightMarginWillChange", data=dict(oldValue=oldValue, newValue=value))
            self.width = xMax + value
            self.dirty = True
            self.postNotification(notification="Glyph.RightMarginDidChange", data=dict(oldValue=oldValue, newValue=value))

    rightMargin = property(_get_rightMargin, _set_rightMargin, doc="The right margin of the glyph. Setting this posts *Glyph.WidthChanged*, *Glyph.RightMarginWillChange*, *Glyph.RightMarginDidChange*  and *Glyph.Changed* notifications among others.")

    def _get_bottomMargin(self):
        bounds = self.bounds
        if bounds is None:
            return None
        xMin, yMin, xMax, yMax = bounds
        if self.verticalOrigin is None:
            return yMin
        else:
            return yMin - (self.verticalOrigin - self.height)

    def _set_bottomMargin(self, value):
        bounds = self.bounds
        if bounds is None:
            return
        xMin, yMin, xMax, yMax = bounds
        if self.verticalOrigin is None:
            oldValue = yMin
            self.verticalOrigin = self.height
        else:
            oldValue = yMin - (self.verticalOrigin - self.height)
        diff = value - oldValue
        if value != oldValue:
            self.postNotification(notification="Glyph.BottomMarginWillChange", data=dict(oldValue=oldValue, newValue=value))
            self.height += diff
            self.dirty = True
            self.postNotification(notification="Glyph.BottomMarginDidChange", data=dict(oldValue=oldValue, newValue=value))

    bottomMargin = property(_get_bottomMargin, _set_bottomMargin, doc="The bottom margin of the glyph. Setting this posts *Glyph.HeightChanged*, *Glyph.BottomMarginWillChange*, *Glyph.BottomMarginDidChange*  and *Glyph.Changed* notifications among others.")

    def _get_topMargin(self):
        bounds = self.bounds
        if bounds is None:
            return None
        xMin, yMin, xMax, yMax = bounds
        if self.verticalOrigin is None:
            return self._height - yMax
        else:
            return self.verticalOrigin - yMax

    def _set_topMargin(self, value):
        bounds = self.bounds
        if bounds is None:
            return
        xMin, yMin, xMax, yMax = bounds
        if self.verticalOrigin is None:
            oldValue = self._height - yMax
        else:
            oldValue = self.verticalOrigin - yMax
        diff = value - oldValue
        if oldValue != value:
            self.postNotification(notification="Glyph.TopMarginWillChange", data=dict(oldValue=oldValue, newValue=value))
            self.verticalOrigin = yMax + value
            self.height += diff
            self.dirty = True
            self.postNotification(notification="Glyph.TopMarginWillChange", data=dict(oldValue=oldValue, newValue=value))

    topMargin = property(_get_topMargin, _set_topMargin, doc="The top margin of the glyph. Setting this posts *Glyph.HeightChanged*, *Glyph.VerticalOriginChanged*, *Glyph.TopMarginWillChange*, *Glyph.TopMarginDidChange*  and *Glyph.Changed* notifications among others.")

    # width

    def _get_width(self):
        return self._width

    def _set_width(self, value):
        oldValue = self._width
        if oldValue != value:
            self._width = value
            self.postNotification(notification="Glyph.WidthChanged", data=dict(oldValue=oldValue, newValue=value))
            self.dirty = True

    width = property(_get_width, _set_width, doc="The width of the glyph. Setting this posts *Glyph.WidthChanged* and *Glyph.Changed* notifications.")

    # height

    def _get_height(self):
        return self._height

    def _set_height(self, value):
        oldValue = self._height
        if oldValue != value:
            self._height = value
            self.postNotification(notification="Glyph.HeightChanged", data=dict(oldValue=oldValue, newValue=value))
            self.dirty = True

    height = property(_get_height, _set_height, doc="The height of the glyph. Setting this posts *Glyph.HeightChanged* and *Glyph.Changed* notifications.")

    # ----------------------
    # Lib Wrapped Attributes
    # ----------------------

    # mark color

    def _get_markColor(self):
        value = self.lib.get("public.markColor")
        if value is not None:
            value = Color(value)
        return value

    def _set_markColor(self, value):
        # convert to a color object
        if value is not None:
            value = Color(value)
        # don't write if there is no change
        oldValue = self.lib.get("public.markColor")
        if oldValue is not None:
            oldValue = Color(oldValue)
        if value == oldValue:
            return
        # remove
        if value is None:
            if "public.markColor" in self.lib:
                del self.lib["public.markColor"]
        # store
        else:
            self.lib["public.markColor"] = value
        self.postNotification(notification="Glyph.MarkColorChanged", data=dict(oldValue=oldValue, newValue=value))

    markColor = property(_get_markColor, _set_markColor, doc="The glyph's mark color. When setting, the value can be a UFO color string, a sequence of (r, g, b, a) or a :class:`Color` object. Setting this posts *Glyph.MarkColorChanged* and *Glyph.Changed* notifications.")

    # vertical origin

    def _get_verticalOrigin(self):
        value = self.lib.get("public.verticalOrigin")
        return value

    def _set_verticalOrigin(self, value):
        # don't write if there is no change
        oldValue = self.lib.get("public.verticalOrigin")
        if value == oldValue:
            return
        # remove
        if value is None:
            if "public.verticalOrigin" in self.lib:
                del self.lib["public.verticalOrigin"]
        # store
        else:
            self.lib["public.verticalOrigin"] = value
        self.postNotification(notification="Glyph.VerticalOriginChanged", data=dict(oldValue=oldValue, newValue=value))

    verticalOrigin = property(_get_verticalOrigin, _set_verticalOrigin, doc="The glyph's vertical origin. Setting this posts *Glyph.VerticalOriginChanged* and *Glyph.Changed* notifications.")

    # -------
    # Pen API
    # -------

    def draw(self, pen):
        """
        Draw the glyph with **pen**.
        """
        from fontTools.pens.pointPen import PointToSegmentPen
        pointPen = PointToSegmentPen(pen)
        self.drawPoints(pointPen)

    def drawPoints(self, pointPen):
        """
        Draw the glyph with **pointPen**.
        """
        if self._shallowLoadedContours:
            self._drawShallowLoadedContours(pointPen, self._shallowLoadedContours)
        else:
            for contour in self._contours:
                contour.drawPoints(pointPen)
        for component in self._components:
            component.drawPoints(pointPen)

    def _drawShallowLoadedContours(self, pointPen, contours):
        for contour in contours:
            try:
                pointPen.beginPath(identifier=contour.get("identifier"))
            except TypeError:
                pointPen.beginPath()
                warn("The beginPath method needs an identifier kwarg. The contour's identifier value has been discarded.", DeprecationWarning)
            for args, kwargs in contour["points"]:
                pointPen.addPoint(*args, **kwargs)
            pointPen.endPath()

    def getPen(self):
        """
        Get the pen used to draw into this glyph.
        """
        from fontTools.pens.pointPen import SegmentToPointPen
        return SegmentToPointPen(self.getPointPen())

    def getPointPen(self):
        """
        Get the point pen used to draw into this glyph.
        """
        from defcon.pens.glyphObjectPointPen import GlyphObjectPointPen, GlyphObjectLoadingPointPen
        if self._isLoading:
            self._shallowLoadedContours = []
            return GlyphObjectLoadingPointPen(self)
        else:
            return GlyphObjectPointPen(self)

    # --------
    # Contours
    # --------

    def _get_contourClass(self):
        return self._contourClass

    contourClass = property(_get_contourClass, doc="The class used for contours.")

    def _get_pointClass(self):
        return self._pointClass

    pointClass = property(_get_pointClass, doc="The class used for points.")

    def _fullyLoadShallowLoadedContours(self):
        if not self._shallowLoadedContours:
            self._shallowLoadedContours = None
            return
        self.disableNotifications()
        contours = list(self._shallowLoadedContours)
        self._shallowLoadedContours = None
        dirty = self.dirty
        pointPen = self.getPointPen()
        self._drawShallowLoadedContours(pointPen, contours)
        self.dirty = dirty
        self.enableNotifications()

    def instantiateContour(self, contourDict=None):
        contour = self._contourClass(
            glyph=self,
            pointClass=self.pointClass
        )
        if contourDict is not None:
            contour.setDataFromSerialization(contourDict)
        return contour

    def beginSelfContourNotificationObservation(self, contour):
        if contour.dispatcher is None:
            return
        contour.addObserver(observer=self, methodName="_contourChanged", notification="Contour.Changed")

    def endSelfContourNotificationObservation(self, contour):
        if contour.dispatcher is None:
            return
        contour.removeObserver(observer=self, notification="Contour.Changed")
        contour.endSelfNotificationObservation()

    def appendContour(self, contour):
        """
        Append **contour** to the glyph. The contour must be a defcon
        :class:`Contour` object or a subclass of that object. An error
        will be raised if the contour's identifier or a point identifier
        conflicts with any of the identifiers within the glyph.

        This will post a *Glyph.Changed* notification.
        """
        self.insertContour(len(self), contour)

    def insertContour(self, index, contour):
        """
        Insert **contour** into the glyph at index. The contour
        must be a defcon :class:`Contour` object or a subclass
        of that object. An error will be raised if the contour's
        identifier or a point identifier conflicts with any of
        the identifiers within the glyph.

        This will post a *Glyph.Changed* notification.
        """
        assert contour not in self
        assert contour.glyph in (self, None), "This contour belongs to another glyph."
        self.postNotification(notification="Glyph.ContourWillBeAdded", data=dict(object=contour))
        if contour.glyph is None:
            identifiers = self._identifiers
            if contour.identifier is not None:
                assert contour.identifier not in identifiers
                identifiers.add(contour.identifier)
            for point in contour:
                if point.identifier is not None:
                    assert point.identifier not in identifiers
                    identifiers.add(point.identifier)
            contour.glyph = self
            contour.beginSelfNotificationObservation()
        self.beginSelfContourNotificationObservation(contour)
        self._contours.insert(index, contour)
        self.postNotification(notification="Glyph.ContoursChanged")
        self.dirty = True

    def removeContour(self, contour):
        """
        Remove **contour** from the glyph.

        This will post a *Glyph.Changed* notification.
        """
        if contour not in self:
            raise IndexError("contour not in glyph")
        self.postNotification(notification="Glyph.ContourWillBeDeleted", data=dict(object=contour))
        identifiers = self._identifiers
        if contour.identifier is not None:
            identifiers.remove(contour.identifier)
        for point in contour:
            if point.identifier is not None:
                identifiers.remove(point.identifier)
        self._contours.remove(contour)
        self.endSelfContourNotificationObservation(contour)
        self.postNotification(notification="Glyph.ContoursChanged")
        self.dirty = True

    def contourIndex(self, contour):
        """
        Get the index for **contour**.
        """
        return self._getContourIndex(contour)

    def clearContours(self):
        """
        Clear all contours from the glyph.

        This posts a *Glyph.Changed* notification.
        """
        self.holdNotifications(note="Requested by Glyph.clearContours.")
        for contour in reversed(self):
            self.removeContour(contour)
        self.releaseHeldNotifications()

    def correctContourDirection(self, trueType=False, segmentLength=10):
        """
        Correct the direction of all contours in the glyph.

        This posts a *Glyph.Changed* notification.
        """
        # set the contours to the same direction
        for contour in self:
            contour.clockwise = False
        # sort the contours by area in reverse (i.e. largest first)
        contours = sorted(self, key=lambda contour: -contour.area)
        # build a tree of nested contours
        tree = {}
        for largeIndex, largeContour in enumerate(contours):
            for smallContour in contours[largeIndex + 1:]:
                if largeContour.contourInside(smallContour, segmentLength=segmentLength):
                    if largeContour not in tree:
                        tree[largeContour] = []
                    tree[largeContour].append(smallContour)
        # run through the tree, largest to smallest, flipping
        # the direction of each contour nested within another contour
        for largeContour in contours:
            if largeContour in tree:
                for smallContour in tree[largeContour]:
                    smallContour.reverse()
        # set to the opposite if needed
        if trueType:
            for contour in self:
                contour.reverse()

    # ----------
    # Components
    # ----------

    def _get_componentClass(self):
        return self._componentClass

    componentClass = property(_get_componentClass, doc="The class used for components.")

    def _get_components(self):
        return list(self._components)

    components = property(_get_components, doc="An ordered list of :class:`Component` objects stored in the glyph.")

    def instantiateComponent(self, componentDict=None):
        component = self._componentClass(
            glyph=self
        )
        if componentDict is not None:
            component.setDataFromSerialization(componentDict)
        return component

    def beginSelfComponentNotificationObservation(self, component):
        if component.dispatcher is None:
            return
        component.addObserver(observer=self, methodName="_componentChanged", notification="Component.Changed")
        component.addObserver(observer=self, methodName="_componentBaseGlyphDataChanged", notification="Component.BaseGlyphDataChanged")

    def endSelfComponentNotificationObservation(self, component):
        if component.dispatcher is None:
            return
        component.removeObserver(observer=self, notification="Component.Changed")
        component.removeObserver(observer=self, notification="Component.BaseGlyphDataChanged")
        component.endSelfNotificationObservation()

    def appendComponent(self, component):
        """
        Append **component** to the glyph. The component must be a defcon
        :class:`Component` object or a subclass of that object. An error
        will be raised if the component's identifier conflicts with any of
        the identifiers within the glyph.

        This will post a *Glyph.Changed* notification.
        """
        self.insertComponent(len(self._components), component)

    def insertComponent(self, index, component):
        """
        Insert **component** into the glyph at index. The component
        must be a defcon :class:`Component` object or a subclass
        of that object. An error will be raised if the component's
        identifier conflicts with any of the identifiers within
        the glyph.

        This will post a *Glyph.Changed* notification.
        """
        assert component not in self._components
        assert component.glyph in (self, None), "This component belongs to another glyph."
        self.postNotification(notification="Glyph.ComponentWillBeAdded", data=dict(object=component))
        if component.glyph is None:
            if component.identifier is not None:
                identifiers = self._identifiers
                assert component.identifier not in identifiers
                identifiers.add(component.identifier)
            component.glyph = self
            component.beginSelfNotificationObservation()
        self.beginSelfComponentNotificationObservation(component)
        self._components.insert(index, component)
        self.postNotification(notification="Glyph.ComponentsChanged")
        self.dirty = True

    def removeComponent(self, component):
        """
        Remove **component** from the glyph.

        This will post a *Glyph.Changed* notification.
        """
        self.postNotification(notification="Glyph.ComponentWillBeDeleted", data=dict(object=component))
        if component.identifier is not None:
            self._identifiers.remove(component.identifier)
        self._components.remove(component)
        self.endSelfComponentNotificationObservation(component)
        self.postNotification(notification="Glyph.ComponentsChanged")
        self.dirty = True

    def componentIndex(self, component):
        """
        Get the index for **component**.
        """
        return self._components.index(component)

    def clearComponents(self):
        """
        Clear all components from the glyph.

        This posts a *Glyph.Changed* notification.
        """
        self.holdNotifications(note="Requested by Glyph.clearComponents.")
        for component in reversed(self._components):
            self.removeComponent(component)
        self.releaseHeldNotifications()

    def decomposeComponent(self, component):
        """
        Decompose **component**. This will preserve the identifiers
        in the incoming contours and points unless there is a conflict.
        In that case, the conflicting incoming identifier will be discarded.

        This posts *Glyph.ComponentsChanged*, *Glyph.ContoursChanged*
        and *Glyph.Changed* notifications.
        """
        self.holdNotifications(note="Requested by Glyph.decomposeComponent.")
        layer = self.layer
        pointPen = DecomposeComponentPointPen(self, layer)
        self._decomposeComponent(component, layer, pointPen)
        self.releaseHeldNotifications()
        self.postNotification(notification="Glyph.ContoursChanged")

    def decomposeAllComponents(self):
        """
        Decompose all components in this glyph. This will preserve the
        identifiers in the incoming contours and points unless there is a
        conflict. In that case, the conflicting incoming identifier will
        be discarded.

        This posts *Glyph.ComponentsChanged*, *Glyph.ContoursChanged*
        and *Glyph.Changed* notifications.
        """
        if not self.components:
            return
        self.holdNotifications(note="Requested by Glyph.decomposeAllComponents.")
        layer = self.layer
        pointPen = DecomposeComponentPointPen(self, layer)
        for component in self.components:
            self._decomposeComponent(component, layer, pointPen)
        self.releaseHeldNotifications()
        self.postNotification(notification="Glyph.ContoursChanged")

    def _decomposeComponent(self, component, layer, pointPen):
        pointPen.skipConflictingIdentifiers = True
        component.drawPoints(pointPen)
        self.removeComponent(component)

    # -------
    # Anchors
    # -------

    def _get_anchorClass(self):
        return self._anchorClass

    anchorClass = property(_get_anchorClass, doc="The class used for anchors.")

    def _get_anchors(self):
        return list(self._anchors)

    def _set_anchors(self, value):
        self.clearAnchors()
        self.holdNotifications(note="Requested by Glyph._set_anchors.")
        for anchor in value:
            self.appendAnchor(anchor)
        self.releaseHeldNotifications()

    anchors = property(_get_anchors, _set_anchors, doc="An ordered list of :class:`Anchor` objects stored in the glyph.")

    def instantiateAnchor(self, anchorDict=None):
        anchor = self._anchorClass(
            glyph=self,
            anchorDict=anchorDict)
        return anchor

    def beginSelfAnchorNotificationObservation(self, anchor):
        if anchor.dispatcher is None:
            return
        anchor.addObserver(observer=self, methodName="_anchorChanged", notification="Anchor.Changed")

    def endSelfAnchorNotificationObservation(self, anchor):
        if anchor.dispatcher is None:
            return
        anchor.removeObserver(observer=self, notification="Anchor.Changed")
        anchor.endSelfNotificationObservation()

    def appendAnchor(self, anchor):
        """
        Append **anchor** to the glyph. The anchor must be a defcon
        :class:`Anchor` object or a subclass of that object. An error
        will be raised if the anchor's identifier conflicts with any of
        the identifiers within the glyph.

        This will post a *Glyph.Changed* notification.
        """
        self.insertAnchor(len(self._anchors), anchor)

    def insertAnchor(self, index, anchor):
        """
        Insert **anchor** into the glyph at index. The anchor
        must be a defcon :class:`Anchor` object or a subclass
        of that object. An error will be raised if the anchor's
        identifier conflicts with any of the identifiers within
        the glyph.

        This will post a *Glyph.Changed* notification.
        """
        if not isinstance(anchor, self._anchorClass):
            anchor = self.instantiateAnchor(anchorDict=anchor)

        assert anchor not in self._anchors
        assert anchor.glyph in (self, None), "This anchor belongs to another glyph."

        self.postNotification(notification="Glyph.AnchorWillBeAdded", data=dict(object=anchor))
        if anchor.glyph is None:
            if anchor.identifier is not None:
                identifiers = self._identifiers
                assert anchor.identifier not in identifiers
                identifiers.add(anchor.identifier)
            anchor.glyph = self
            anchor.beginSelfNotificationObservation()
        self.beginSelfAnchorNotificationObservation(anchor)
        self._anchors.insert(index, anchor)
        self.postNotification(notification="Glyph.AnchorsChanged")
        self.dirty = True

    def removeAnchor(self, anchor):
        """
        Remove **anchor** from the glyph.

        This will post a *Glyph.Changed* notification.
        """
        self.postNotification(notification="Glyph.AnchorWillBeDeleted", data=dict(object=anchor))
        if anchor.identifier is not None:
            self._identifiers.remove(anchor.identifier)
        self._anchors.remove(anchor)
        self.endSelfAnchorNotificationObservation(anchor)
        self.postNotification(notification="Glyph.AnchorsChanged")
        self.dirty = True

    def anchorIndex(self, anchor):
        """
        Get the index for **anchor**.
        """
        return self._anchors.index(anchor)

    def clearAnchors(self):
        """
        Clear all anchors from the glyph.

        This posts a *Glyph.Changed* notification.
        """
        self.holdNotifications(note="Requested by Glyph.clearAnchors.")
        for anchor in reversed(self._anchors):
            self.removeAnchor(anchor)
        self.releaseHeldNotifications()

    # ----------
    # Guidelines
    # ----------

    def _get_guidelineClass(self):
        return self._guidelineClass

    guidelineClass = property(_get_guidelineClass, doc="The class used for guidelines.")

    def _get_guidelines(self):
        return list(self._guidelines)

    def _set_guidelines(self, value):
        self.clearGuidelines()
        self.holdNotifications(note="Requested by Glyph._set_guidelines.")
        for guideline in value:
            self.appendGuideline(guideline)
        self.releaseHeldNotifications()

    guidelines = property(_get_guidelines, _set_guidelines, doc="An ordered list of :class:`Guideline` objects stored in the glyph. Setting this will post a *Glyph.Changed* notification along with any notifications posted by the :py:meth:`Glyph.appendGuideline` and :py:meth:`Glyph.clearGuidelines` methods.")

    def instantiateGuideline(self, guidelineDict=None):
        guideline = self._guidelineClass(
            glyph=self,
            guidelineDict=guidelineDict
        )
        return guideline

    def beginSelfGuidelineNotificationObservation(self, guideline):
        if guideline.dispatcher is None:
            return
        guideline.addObserver(observer=self, methodName="_guidelineChanged", notification="Guideline.Changed")

    def endSelfGuidelineNotificationObservation(self, guideline):
        if guideline.dispatcher is None:
            return
        guideline.removeObserver(observer=self, notification="Guideline.Changed")
        guideline.endSelfNotificationObservation()

    def appendGuideline(self, guideline):
        """
        Append **guideline** to the glyph. The guideline must be a defcon
        :class:`Guideline` object or a subclass of that object. An error
        will be raised if the guideline's identifier conflicts with any of
        the identifiers within the glyph.

        This will post a *Glyph.Changed* notification.
        """
        self.insertGuideline(len(self._guidelines), guideline)

    def insertGuideline(self, index, guideline):
        """
        Insert **guideline** into the glyph at index. The guideline
        must be a defcon :class:`Guideline` object or a subclass
        of that object. An error will be raised if the guideline's
        identifier conflicts with any of the identifiers within
        the glyph.

        This will post a *Glyph.Changed* notification.
        """

        assert guideline not in self.guidelines
        if not isinstance(guideline, self._guidelineClass):
            guideline = self.instantiateGuideline(guidelineDict=guideline)

        assert guideline.glyph in (self, None), "This guideline belongs to another glyph."
        if guideline.glyph is None:
            assert guideline.font is None, "This guideline belongs to a font."
        self.postNotification(notification="Glyph.GuidelineWillBeAdded", data=dict(object=guideline))
        if guideline.glyph is None:
            if guideline.identifier is not None:
                identifiers = self._identifiers
                assert guideline.identifier not in identifiers
                if guideline.identifier is not None:
                    identifiers.add(guideline.identifier)
            guideline.glyph = self
            guideline.beginSelfNotificationObservation()
        self.beginSelfGuidelineNotificationObservation(guideline)
        self._guidelines.insert(index, guideline)
        self.postNotification(notification="Glyph.GuidelinesChanged")
        self.dirty = True

    def removeGuideline(self, guideline):
        """
        Remove **guideline** from the glyph.

        This will post a *Glyph.Changed* notification.
        """
        self.postNotification(notification="Glyph.GuidelineWillBeDeleted", data=dict(object=guideline))
        if guideline.identifier is not None:
            self._identifiers.remove(guideline.identifier)
        self._guidelines.remove(guideline)
        self.endSelfGuidelineNotificationObservation(guideline)
        self.postNotification(notification="Glyph.GuidelinesChanged")
        self.dirty = True

    def guidelineIndex(self, guideline):
        """
        Get the index for **guideline**.
        """
        return self._guidelines.index(guideline)

    def clearGuidelines(self):
        """
        Clear all guidelines from the glyph.

        This posts a *Glyph.Changed* notification.
        """
        self.holdNotifications(note="Requested by Glyph.clearGuidelines.")
        for guideline in reversed(self._guidelines):
            self.removeGuideline(guideline)
        self.releaseHeldNotifications()

    # ----
    # Note
    # ----

    def _get_note(self):
        return self._note

    def _set_note(self, value):
        if value is not None:
            assert isinstance(value, str)
        oldValue = self._note
        if oldValue != value:
            self._note = value
            self.postNotification(notification="Glyph.NoteChanged", data=dict(oldValue=oldValue, newValue=value))
            self.dirty = True

    note = property(_get_note, _set_note, doc="An arbitrary note for the glyph. Setting this will post a *Glyph.Changed* notification.")

    # ---
    # Lib
    # ---

    def _get_libClass(self):
        return self._libClass

    libClass = property(_get_libClass, doc="The class used for the lib.")

    def instantiateLib(self):
        lib = self._libClass(
            glyph=self
        )
        return lib

    def _get_lib(self):
        if self._lib is None:
            self._lib = self.instantiateLib()
            self.beginSelfLibNotificationObservation()
        return self._lib

    def _set_lib(self, value):
        lib = self.lib
        lib.clear()
        lib.update(value)
        self.dirty = True

    lib = property(_get_lib, _set_lib, doc="The glyph's :class:`Lib` object. Setting this will clear any existing lib data and post a *Glyph.Changed* notification if data was replaced.")

    def beginSelfLibNotificationObservation(self):
        if self._lib.dispatcher is None:
            return
        self._lib.addObserver(observer=self, methodName="_libContentChanged", notification="Lib.Changed")

    def endSelfLibNotificationObservation(self):
        if self._lib is None:
            return
        if self._lib.dispatcher is None:
            return
        self._lib.removeObserver(observer=self, notification="Lib.Changed")
        self._lib.endSelfNotificationObservation()

    # temp lib

    def _get_tempLib(self):
        if self._tempLib is None:
            self._tempLib = self.instantiateLib()
        return self._tempLib

    def _set_tempLib(self, value):
        if value is not None:
            self.tempLib.clear()
            self.tempLib.update(value)

    tempLib = property(_get_tempLib, _set_tempLib, doc="The glyph's :class:`tempLib` object.")

    # -----
    # Image
    # -----

    def _get_imageClass(self):
        return self._imageClass

    imageClass = property(_get_imageClass, doc="The class used for the image.")

    def instantiateImage(self, imageDict=None):
        image = self._imageClass(
            glyph=self,
            imageDict=imageDict
        )
        return image

    def _get_image(self):
        if self._image is None:
            self._image = self.instantiateImage()
            self.beginSelfImageNotificationObservation()
        return self._image

    def _set_image(self, image):
        # clearing the image
        if image is None:
            if self._image is not None:
                self.postNotification(notification="Glyph.ImageWillBeCleared")
                self._image.clear()
                self.postNotification(notification="Glyph.ImageCleared")
                self.postNotification(notification="Glyph.ImageChanged")
                self.dirty = True
        # adding image
        else:
            if self._image is None:
                # create the image object
                i = self.image
            if set(self._image.items()) != set(image.items()):
                self._image.fileName = image["fileName"]
                self._image.transformation = (image["xScale"], image["xyScale"], image["yxScale"], image["yScale"], image["xOffset"], image["yOffset"])
                self._image.color = image.get("color")
                self.postNotification(notification="Glyph.ImageChanged")
                self.dirty = True

    image = property(_get_image, _set_image, doc="The glyph's :class:`Image` object. Setting this posts *Glyph.ImageChanged* and *Glyph.Changed* notifications.")

    def clearImage(self):
        self.image = None

    def beginSelfImageNotificationObservation(self):
        if self._image.dispatcher is None:
            return
        self._image.addObserver(observer=self, methodName="_imageChanged", notification="Image.Changed")
        self._image.addObserver(observer=self, methodName="_imageDataChanged", notification="Image.ImageDataChanged")

    def endSelfImageNotificationObservation(self):
        if self._image is None:
            return
        if self._image.dispatcher is None:
            return
        self._image.removeObserver(observer=self, notification="Image.Changed")
        self._image.removeObserver(observer=self, notification="Image.ImageDataChanged")
        self._image.endSelfNotificationObservation()

    # -------------
    # List Behavior
    # -------------

    def __contains__(self, contour):
        if self._shallowLoadedContours is not None:
            self._fullyLoadShallowLoadedContours()
        return contour in self._contours

    def __len__(self):
        if self._shallowLoadedContours is not None:
            self._fullyLoadShallowLoadedContours()
        return len(self._contours)

    def __iter__(self):
        if self._shallowLoadedContours is not None:
            self._fullyLoadShallowLoadedContours()
        return iter(self._contours)

    def __getitem__(self, index):
        if self._shallowLoadedContours is not None:
            self._fullyLoadShallowLoadedContours()
        return self._contours[index]

    def _getContourIndex(self, contour):
        if self._shallowLoadedContours is not None:
            self._fullyLoadShallowLoadedContours()
        return self._contours.index(contour)

    # ----------------
    # Glyph Absorption
    # ----------------

    def copyDataFromGlyph(self, glyph):
        """
        Copy data from **glyph**. This copies the following data:

        ==========
        width
        height
        unicodes
        note
        image
        contours
        components
        anchors
        guidelines
        lib
        ==========

        The name attribute is purposefully omitted.
        """
        from copy import deepcopy
        self.width = glyph.width
        self.height = glyph.height
        self.unicodes = list(glyph.unicodes)
        self.note = glyph.note
        self.guidelines = [self.instantiateGuideline(g) for g in glyph.guidelines]
        self.anchors = [self.instantiateAnchor(a) for a in glyph.anchors]
        self.image = glyph.image
        pointPen = self.getPointPen()
        glyph.drawPoints(pointPen)
        self.lib = deepcopy(glyph.lib)

    # -----
    # Clear
    # -----

    def clear(self):
        """
        Clear all contours, components, anchors and guidelines from the glyph.

        This posts a *Glyph.Changed* notification.
        """
        self.holdNotifications(note="Requested by Glyph.clear.")
        self.clearContours()
        self.clearComponents()
        self.clearAnchors()
        self.clearGuidelines()
        self.clearImage()
        self.releaseHeldNotifications()

    # ----
    # Move
    # ----

    def move(self, values):
        """
        Move all contours, components and anchors in the glyph
        by **(x, y)**.

        This posts a *Glyph.Changed* notification.
        """
        (x, y) = values
        for contour in self:
            contour.move((x, y))
        for component in self._components:
            component.move((x, y))
        for anchor in self._anchors:
            anchor.move((x, y))

    # ------------
    # Point Inside
    # ------------

    def pointInside(self, coordinates, evenOdd=False):
        """
        Returns a boolean indicating if **(x, y)** is in the
        "black" area of the glyph.
        """
        (x, y) = coordinates
        from fontTools.pens.pointInsidePen import PointInsidePen
        pen = PointInsidePen(glyphSet=None, testPoint=(x, y), evenOdd=evenOdd)
        self.draw(pen)
        return pen.getResult()

    # ----------------------
    # Notification Callbacks
    # ----------------------

    def endSelfNotificationObservation(self):
        if self.dispatcher is None:
            return
        if self._contours:
            for contour in self:
                self.endSelfContourNotificationObservation(contour)
        for component in self.components:
            self.endSelfComponentNotificationObservation(component)
        for anchor in self.anchors:
            self.endSelfAnchorNotificationObservation(anchor)
        for guideline in self.guidelines:
            self.endSelfGuidelineNotificationObservation(guideline)
        self.endSelfLibNotificationObservation()
        self.endSelfImageNotificationObservation()
        super(Glyph, self).endSelfNotificationObservation()
        self._font = None
        self._layerSet = None
        self._layer = None

    def _imageDataChanged(self, notification):
        self.postNotification(notification="Glyph.ImageChanged")
        self.postNotification(notification=self.changeNotificationName)

    def _imageChanged(self, notification):
        self.postNotification(notification="Glyph.ImageChanged")
        self.dirty = True

    def _contourChanged(self, notification):
        self.postNotification(notification="Glyph.ContoursChanged")
        self.dirty = True

    def _componentChanged(self, notification):
        self.postNotification(notification="Glyph.ComponentsChanged")
        self.dirty = True

    def _componentBaseGlyphDataChanged(self, notification):
        self.postNotification(notification="Glyph.ComponentsChanged")
        self.postNotification(notification=self.changeNotificationName)

    def _anchorChanged(self, notification):
        self.postNotification(notification="Glyph.AnchorsChanged")
        self.dirty = True

    def _guidelineChanged(self, notification):
        self.postNotification(notification="Glyph.GuidelinesChanged")
        self.dirty = True

    def _libContentChanged(self, notification):
        self.postNotification(notification="Glyph.LibChanged")
        self.dirty = True

    # -----------------------------
    # Serialization/Deserialization
    # -----------------------------

    def getDataForSerialization(self, **kwargs):
        from functools import partial

        simple_get = partial(getattr, self)
        serialize = lambda item: item.getDataForSerialization()
        serialized_get = lambda key: serialize(simple_get(key))
        serialized_list_get = lambda key: [serialize(item) for item in simple_get(key)]

        getters = [
            ('name', simple_get),
            ('unicodes', simple_get),
            ('width', simple_get),
            ('height', simple_get),
            ('note', simple_get),
            ('components', serialized_list_get),
            ('anchors', serialized_list_get),
            ('guidelines', serialized_list_get),
            ('image', serialized_get),
            ('lib', serialized_get),
            ('tempLib', serialized_get),
        ]

        if self._shallowLoadedContours is not None:
            getters.append(('_shallowLoadedContours', simple_get))
        else:
            getters.append(('_contours', serialized_list_get))

        return self._serialize(getters, **kwargs)

    def setDataFromSerialization(self, data):
        from functools import partial

        set_attr = partial(setattr, self) # key, data

        def set_each(setter, drop_key=False):

            _setter = lambda k, v: setter(v) if drop_key else setter

            def wrapper(key, data):
                for d in data:
                    _setter(key, d)
            return wrapper

        def single_init(factory, data):
            item = factory(data)
            return item

        def list_init(factory, data):
            return [single_init(factory, childData) for childData in data]

        def init_set(init, factory, setter):
            def wrapper(key, data):
                setter(key, init(factory, data))
            return wrapper

        # Clear all contours, components, anchors and guidelines from the glyph.
        self.clear()

        setters = (
            ('name', set_attr),
            ('unicodes', set_attr),
            ('width', set_attr),
            ('height', set_attr),
            ('note', set_attr),
            ('lib', set_attr),
            ('tempLib', set_attr),
            ('_shallowLoadedContours', set_attr),
            ('_contours', init_set(list_init, self.instantiateContour, set_each(self.appendContour, True))),
            ('components', init_set(list_init, self.instantiateComponent, set_each(self.appendComponent, True))),
            ('guidelines', init_set(list_init, self.instantiateGuideline, set_attr)),
            ('anchors', init_set(list_init, self.instantiateAnchor, set_attr)),
            ('image', init_set(single_init, self.instantiateImage, set_attr))
        )

        for key, setter in setters:
            if key not in data:
                continue
            setter(key, data[key])


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import unittest
from defcon import Font


class GroupsTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_side1KerningGroups(self):
        font = Font()
        groups = font.groups
        groups["a"] = ["a"]
        result = groups.getRepresentation("defcon.groups.kerningSide1Groups")
        self.assertEqual(result, {})
        groups["public.kern1.a"] = ["a"]
        result = groups.getRepresentation("defcon.groups.kerningSide1Groups")
        self.assertEqual(result, {"public.kern1.a" : ["a"]})
        groups["public.kern1.a"] = ["b"]
        result = groups.getRepresentation("defcon.groups.kerningSide1Groups")
        self.assertEqual(result, {"public.kern1.a" : ["b"]})
        groups["public.kern2.a"] = ["b"]
        result = groups.getRepresentation("defcon.groups.kerningSide1Groups")
        self.assertEqual(result, {"public.kern1.a" : ["b"]})

    def test_side2KerningGroups(self):
        font = Font()
        groups = font.groups
        groups["a"] = ["a"]
        result = groups.getRepresentation("defcon.groups.kerningSide2Groups")
        self.assertEqual(result, {})
        groups["public.kern2.a"] = ["a"]
        result = groups.getRepresentation("defcon.groups.kerningSide2Groups")
        self.assertEqual(result, {"public.kern2.a" : ["a"]})
        groups["public.kern2.a"] = ["b"]
        result = groups.getRepresentation("defcon.groups.kerningSide2Groups")
        self.assertEqual(result, {"public.kern2.a" : ["b"]})
        groups["public.kern1.a"] = ["b"]
        result = groups.getRepresentation("defcon.groups.kerningSide2Groups")
        self.assertEqual(result, {"public.kern2.a" : ["b"]})

    def test_side1glyphToKerningGroups(self):
        font = Font()
        groups = font.groups
        groups["public.kern1.a"] = ["a"]
        result = groups.getRepresentation("defcon.groups.kerningGlyphToSide1Group")
        self.assertEqual(result, {'a': 'public.kern1.a'})
        del groups["public.kern1.a"]
        groups["public.kern1.b"] = ["a"]
        result = groups.getRepresentation("defcon.groups.kerningGlyphToSide1Group")
        self.assertEqual(result, {'a': 'public.kern1.b'})

    def test_side2glyphToKerningGroups(self):
        font = Font()
        groups = font.groups
        groups["public.kern2.a"] = ["a"]
        result = groups.getRepresentation("defcon.groups.kerningGlyphToSide2Group")
        self.assertEqual(result, {'a': 'public.kern2.a'})
        del groups["public.kern2.a"]
        groups["public.kern2.b"] = ["a"]
        result = groups.getRepresentation("defcon.groups.kerningGlyphToSide2Group")
        self.assertEqual(result, {'a': 'public.kern2.b'})

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import unittest
from defcon.tools.unicodeTools import (
    decompositionBase, openRelative, closeRelative, category, script, block
)


class UnicodeToolsTest(unittest.TestCase):

    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)

    def test_decompositionBase(self):
        self.assertEqual(decompositionBase(ord("ç")), ord("c"))
        self.assertEqual(decompositionBase(ord("a")), -1)
        self.assertEqual(decompositionBase(0x0001000), -1)

    def test_openRelative(self):
        self.assertEqual(openRelative(ord(")")), ord("("))
        self.assertIsNone(openRelative(ord("a")))
        self.assertIsNone(openRelative(0x0001000))

    def test_closeRelative(self):
        self.assertEqual(closeRelative(ord("(")), ord(")"))
        self.assertIsNone(closeRelative(ord("a")))
        self.assertIsNone(closeRelative(0x0001000))

    def test_category(self):
        self.assertEqual(category(ord("a")), "Ll")
        self.assertEqual(category(0x0001000), "Lo")

    def test_script(self):
        self.assertEqual(script(ord("a")), "Latin")
        self.assertEqual(script(0x00010000), "Linear B")

    def test_block(self):
        self.assertEqual(block(ord("a")), "Basic Latin")
        self.assertEqual(block(0x00010000), "Linear B Syllabary")
